// goconv -- the Go language version of our convective model
package main

// packages needed
import (
	"fmt"
	"os"
	"math"
	"math/rand"
	"gocandis"
	"gompi"
	"params"
	"thermo"
	"toyrad"
)

// global declarations
var mx, my, mz, gx, gy, nx, ny, nz, nt, nj int
var lx, ly, lz, ilx, ily, ilz, nfields, nfields2 int
var srcfields, meanfields, sgsize, msize, msize2 int
var nsize, nsize2, lsize, ilsize, fsize, ssize, issize int

// fixed constants
const (
	
	// accel of gravity (m/s^2)
	Gee = 9.8

	// math constants
	Pi = 3.14159

	// model version
	VERSION = "079"
)

// *******************************************************************
// *******************************************************************
// *******************************************************************
// *******************************************************************
// *******************************************************************
// model entry point
func main() {

	// get command line arguments
	args := os.Args
	if len(args) != 2 {
		fmt.Println("Usage: goconv paramfile")
		os.Exit(1)
	}

	// get the parameter list
	pfile := args[1]
	paramlist := make(map[string]string)
	params.Getparams(pfile, paramlist)

	// start MPI
	comm, nprocs, rank := gompi.Init()

	// initialize data structures
	l := lfields{}
	g := gfields{}
	wtg := wtgfields{}
		
	// initialize local and global fields
	initfields(&l,&g,&wtg,paramlist)

	// get parameters
	it0 := params.Iparam(paramlist,"it0")

	// do initial write
	if rank == 0 && it0 == 0 {
		docandis(pfile,nx,ny,nz,0,&g,&wtg,paramlist)
	}

	// consistency checks
	if gx < 1 || gy < 1 || nprocs != gx*gy {
		panic("gx, gy don't match nprocs")
	}

	// scatter global field data to local fields
	scatter(rank,gx,gy,nx,ny,nz,&l,&g)

	// display loop
	for it := 0; it < nt; it++ {
		
		// step in time -- includes both coarse and fine
		// scale time steps
		timestep(comm,rank,nx,ny,nz,it,&l,&g,&wtg,paramlist)

		// step write after gathering data from all processes
		gather(comm,rank,gx,gy,ny,nz,&l,&g)
		if rank == 0 && it >= it0 {
			docandis(pfile,nx,ny,nz,it + 1,&g,&wtg,paramlist)
		}
	}

	// finish MPI
	gompi.Finalize()
}

// *******************************************************************
// *******************************************************************
// *******************************************************************
// *******************************************************************
// *******************************************************************
// initialize global fields
func initfields(l *lfields, g *gfields, wtg *wtgfields,
	paramlist map[string]string) {

	// initialize thermodynamic tables
	ice := params.Iparam(paramlist,"ice")
	thermo.Init(ice)

	// get initialization parameters
	mx = params.Iparam(paramlist,"mx")  // x size local domain
	my = params.Iparam(paramlist,"my")  // y size local domain
	mz = params.Iparam(paramlist,"mz")  // z size
	gx = params.Iparam(paramlist,"gx")  // x local domains
	gy = params.Iparam(paramlist,"gy")  // y local domains
	nt = params.Iparam(paramlist,"nt")  // output loops
	dx := params.Fparam(paramlist,"dx")
	dy := params.Fparam(paramlist,"dy")
	dz := params.Fparam(paramlist,"dz")
	tp := params.Fparam(paramlist,"tp")
	tlapse := params.Fparam(paramlist,"tlapse")
	rh := params.Fparam(paramlist,"rh")
	tsurf := params.Fparam(paramlist,"tsurf")
	vx0 := params.Fparam(paramlist,"vx0")
	vy0 := params.Fparam(paramlist,"vy0")
	sx0 := params.Fparam(paramlist,"sx0")
	sy0 := params.Fparam(paramlist,"sy0")
	cx0 := params.Fparam(paramlist,"cx0")
	cy0 := params.Fparam(paramlist,"cy0")
	ranent0 := params.Fparam(paramlist,"ranent0")
	pointent0 := params.Fparam(paramlist,"pointent0")
	ranseed := params.Iparam(paramlist,"ranseed")
	zscale := params.Fparam(paramlist,"zscale")
	rscale := params.Fparam(paramlist,"rscale")
	mumax := params.Fparam(paramlist,"mumax")
	initfile := params.Sparam(paramlist,"initfile")
	sstparam := params.Fparam(paramlist,"sst0")

	// needed for indexing
	lx = mx + 2
	ly = my + 2
	lz = mz
	ilx = mx + 1
	ily = my + 1
	ilz = mz + 1
	nx = mx*gx
	ny = my*gy
	nz = mz

	// number of dynamical core fields
	nfields = 18

	// number of 2D fields
	nfields2 = 10

	// field numbers
	srcfields = 4
	meanfields = 6

	// size of scatter/gather buffers
	sgsize = nfields*mx*my*mz + nfields2*mx*my
	
	// local segment of global fields
	msize = mx*my*mz
	msize2 = mx*my

	// sizes of global fields
	nsize = nx*ny*nz
	nsize2 = nx*ny

	// extended size of unstaggered local grid
	lsize = lx*ly*lz

	// size of staggered local grid
	ilsize = ilx*ily*ilz

	// size for flux buffers
	// (staggered and unstaggered)
	fsize = lx*ly*ilz

	// surface grid size (unstaggered)
	ssize = lx*ly

	// surface grid size (staggered)
	issize = ilx*ily

	// allocate space for local unstaggered grid fields
	l.rho = make([]float64,lsize)
	l.ux = make([]float64,lsize)
	l.uy = make([]float64,lsize)
	l.uz = make([]float64,lsize)
	l.dent = make([]float64,lsize)
	l.p = make([]float64,lsize)
	l.drt = make([]float64,lsize)
	l.drr = make([]float64,lsize)
	l.keddy = make([]float64,lsize)

	// local staggered grid fields
	l.irho = make([]float64,ilsize)
	l.iux = make([]float64,ilsize)
	l.iuy = make([]float64,ilsize)
	l.iuz = make([]float64,ilsize)
	l.ident = make([]float64,ilsize)
	l.ip = make([]float64,ilsize)
	l.idrt = make([]float64,ilsize)
	l.idrr = make([]float64,ilsize)
	l.ikeddy = make([]float64,ilsize)

	// local staggered 1D fields
	l.entdrrflux = make([]float64,ilz)

	// source terms on unstaggered grid
	l.sux = make([]float64,lsize)
	l.suy = make([]float64,lsize)
	l.suz = make([]float64,lsize)
	l.sdent = make([]float64,lsize)
	l.sdrt = make([]float64,lsize)
	l.sdrr = make([]float64,lsize)
	l.sdrad = make([]float64,lsize)
	l.sflux = make([]float64,lsize)
	l.flux = make([]float64,lsize)

	// unstaggered grids divided by density
	l.vx = make([]float64,lsize)
	l.vy = make([]float64,lsize)
	l.vz = make([]float64,lsize)
	l.ent = make([]float64,lsize)
	l.rt = make([]float64,lsize)
	l.rr = make([]float64,lsize)

	// flux buffers (staggered and unstaggered)
	l.fxx = make([]float64,fsize)
	l.fyy = make([]float64,fsize)
	l.fzz = make([]float64,fsize)
	l.fxy = make([]float64,fsize)
	l.fyz = make([]float64,fsize)
	l.fzx = make([]float64,fsize)
	l.fxent = make([]float64,fsize)
	l.fyent = make([]float64,fsize)
	l.fzent = make([]float64,fsize)
	l.fxrt = make([]float64,fsize)
	l.fyrt = make([]float64,fsize)
	l.fzrt = make([]float64,fsize)
	l.fxrr = make([]float64,fsize)
	l.fyrr = make([]float64,fsize)
	l.fzrr = make([]float64,fsize)

	// coarse timestep flux buffers
	l.Fvxx = make([]float64,ilsize)
	l.Fvyy = make([]float64,ilsize)
	l.Fvzz = make([]float64,ilsize)
	l.Fvxy = make([]float64,ilsize)
	l.Fvyz = make([]float64,ilsize)
	l.Fvzx = make([]float64,ilsize)
	l.Fentx = make([]float64,ilsize)
	l.Fenty = make([]float64,ilsize)
	l.Fentz = make([]float64,ilsize)
	l.Frtx = make([]float64,ilsize)
	l.Frty = make([]float64,ilsize)
	l.Frtz = make([]float64,ilsize)
	l.Frrx = make([]float64,ilsize)
	l.Frry = make([]float64,ilsize)
	l.Frrz = make([]float64,ilsize)

	// also need dry entropy on unstaggered grid
	// this is computed in coarse() and is needed
	// in wtgcalc()
	l.dryent = make([]float64,lsize)
	l.ddryent = make([]float64,lsize)

	// surface fields on unstaggered grid
	l.surfvx = make([]float64,ssize)
	l.surfvy = make([]float64,ssize)
	l.surfent = make([]float64,ssize)
	l.surfrt = make([]float64,ssize)
	l.rain = make([]float64,ssize)
	l.surfentcum = make([]float64,ssize)
	l.surfrtcum = make([]float64,ssize)
	l.raincum = make([]float64,ssize)
	l.sst = make([]float64,ssize)
	l.surft = make([]float64,ssize)

	// surface fields on staggered grid
	l.isurfvx = make([]float64,issize)
	l.isurfvy = make([]float64,issize)
	l.isurfent = make([]float64,issize)
	l.isurfrt = make([]float64,issize)
	l.irain = make([]float64,issize)
	l.isst = make([]float64,issize)
	l.isurft = make([]float64,issize)

	// buffers for MPI transfers so that
	// processors can talk to nearest neighbors
	l.sbuffn = make([]float64,nfields*(mx + 2)*mz)
	l.rbuffn = make([]float64,nfields*(mx + 2)*mz)
	l.sbuffs = make([]float64,nfields*(mx + 2)*mz)
	l.rbuffs = make([]float64,nfields*(mx + 2)*mz)
	l.sbuffe = make([]float64,nfields*(my + 2)*mz)
	l.rbuffe = make([]float64,nfields*(my + 2)*mz)
	l.sbuffw = make([]float64,nfields*(my + 2)*mz)
	l.rbuffw = make([]float64,nfields*(my + 2)*mz)

	// scatter/gather buffers
	l.buff = make([]float64,sgsize)
	l.rbuff = make([]float64,sgsize)

	// global fields
	g.x = make([]float64,nx)
	g.y = make([]float64,ny)
	g.z = make([]float64,nz)
	g.rho = make([]float64,nsize)
	g.ux = make([]float64,nsize)
	g.uy = make([]float64,nsize)
	g.uz = make([]float64,nsize)
	g.dent = make([]float64,nsize)
	g.p = make([]float64,nsize)
	g.drt = make([]float64,nsize)
	g.drr = make([]float64,nsize)
	g.keddy = make([]float64,nsize)
	g.sux = make([]float64,nsize)
	g.suy = make([]float64,nsize)
	g.suz = make([]float64,nsize)
	g.sdent = make([]float64,nsize)
	g.sdrt = make([]float64,nsize)
	g.sdrr = make([]float64,nsize)
	g.sdrad = make([]float64,nsize)
	g.sflux = make([]float64,nsize)
	g.flux = make([]float64,nsize)
	g.surfvx = make([]float64,nsize2)
	g.surfvy = make([]float64,nsize2)
	g.surfent = make([]float64,nsize2)
	g.surfrt = make([]float64,nsize2)
	g.rain = make([]float64,nsize2)
	g.surfentcum = make([]float64,nsize2)
	g.surfrtcum = make([]float64,nsize2)
	g.raincum = make([]float64,nsize2)
	g.sst = make([]float64,nsize2)
	g.surft = make([]float64,nsize2)

	// calculate number of spectral modes for wtg
	nj = int(tp/dz + 0.5)

	// allocate space for wtg fields
	wtg.ux = make([]float64,nz)
	wtg.uy = make([]float64,nz)
	wtg.pres = make([]float64,nz)
	wtg.rho = make([]float64,nz)
	wtg.drt = make([]float64,nz)
	wtg.ddryent = make([]float64,nz)
	wtg.dent = make([]float64,nz)
	wtg.dsatent = make([]float64,nz)
	wtg.uz = make([]float64,nz)

	wtg.uxmean = make([]float64,nz)
	wtg.uymean = make([]float64,nz)
	wtg.rhomean = make([]float64,nz)
	wtg.drtmean = make([]float64,nz)
	wtg.ddryentmean = make([]float64,nz)
	wtg.dentmean = make([]float64,nz)

	wtg.srcvx = make([]float64,nz)
	wtg.srcvy = make([]float64,nz)
	wtg.srcrt = make([]float64,nz)
	wtg.srcent = make([]float64,nz)
	wtg.srcrtcum = make([]float64,nz)
	wtg.srcentcum = make([]float64,nz)

	wtg.dryentj = make([]float64,nj)
	wtg.sine0 = make([]float64,nz*nj)
	wtg.mu = make([]float64,nz)

	wtg.sgbuff = make([]float64,meanfields*nz)
	wtg.rgbuff = make([]float64,meanfields*nz*gx*gy)
	wtg.ssbuff = make([]float64,srcfields*nz)
	wtg.rsbuff = make([]float64,srcfields*nz)

	// do thermodynamic profile initialization
	if initfile == "-" {

		// extrapolate temp and pres to first model level
		// using hydrostatic eqn
		// the reference pressure is the surface pressure
		// assume dry adiabat for temperature
		dhalf := dz/2.0
		temp0 := tsurf - (Gee/thermo.Cp)*dhalf
		tstar := 0.5*(tsurf + temp0)
		pstar := thermo.Pref - 0.5*Gee*thermo.Pref*dhalf/(thermo.R*tstar)
		p0 := thermo.Pref - Gee*pstar*dhalf/(thermo.R*tstar)

		// get SST from paramfile value
		l.sst0 = sstparam

		// zero the domain propagation velocity
		wtg.vxsys = 0.0
		wtg.vysys = 0.0

		// protect against negative RH
		if rh < 0.0 {
			fmt.Println("goconv: RH < 0")
			os.Exit(1)
		}

		// compute hydrostatic profile
		wtg.pres, wtg.rho, wtg.drt, wtg.ddryent, wtg.dent,
		wtg.dsatent = hydrostatic(nz,temp0,rh,p0,tp,tlapse,dz)
	} else {

		// get profile from initialization file
		wtg.pres, wtg.rho, wtg.drt, wtg.ddryent, wtg.dent,
		wtg.dsatent, wtg.ux, wtg.uy,
		wtg.vxsys, wtg.vysys, l.sst0 = getinit(nz,initfile)
	}

	// initialize wtg velocity fields
	for iz := 0; iz < nz; iz++ {
		zscaled := float64(iz) + 0.5
		z := dz*zscaled

		// make wind constant above tropopause
		if z > tp {z = tp}

		// only do this initialization if no ref profile
		// the domain propagation velocity is subtracted here
		if initfile == "-" {
			wtg.ux[iz] = wtg.rho[iz]*(vx0 + sx0*z + 0.5*cx0*z*z)
			wtg.uy[iz] = wtg.rho[iz]*(vy0 + sy0*z + 0.5*cy0*z*z)
		}

		// normal modes for wtg
		for j := 0; j < nj; j++ {
			wtg.sine0[SI(iz,j)] = math.Sin(Pi*float64(j)*zscaled/
				float64(nj))
		}

		// cumulative WTG source profiles
		wtg.srcentcum[iz] = 0.0
		wtg.srcrtcum[iz] = 0.0
	}

	// initialize stratospheric damping profile
	for iz := 0; iz < nz; iz++ {
		z := dz*(float64(iz) + 0.5)
		ztop := dz*float64(nz)
		wtg.mu[iz] = mumax*(z - tp)/(ztop - tp)
		if wtg.mu[iz] < 0.0 {wtg.mu[iz] = 0.0}
	}

	// initialize global index fields
	for ix := 0; ix < nx; ix++ {
		g.x[ix] = float64(ix)*dx
	}
	for iy := 0; iy < ny; iy++ {
		g.y[iy] = float64(iy)*dy
	}
	for iz := 0; iz < nz; iz++ {
		g.z[iz] = (float64(iz) + 0.5)*dz
	}
	
	// provide a seed so that all processes start
	// the the same seed -- specifiable in input file
	rand.Seed(int64(ranseed))
	for ix := 0; ix < nx; ix++ {
		for iy := 0; iy < ny; iy++ {

			// compute radius from center
			delx := dx*float64(ix - nx/2)
			dely := dy*float64(iy - ny/2)
			radsq := delx*delx + dely*dely
			pointamp := pointent0*math.Exp(-radsq/(rscale*rscale))

			// define initial random perturbation magnitude
			rentinit := ranent0*(2.0*rand.Float64() - 1.0)

			// initialize global 3D fields from wtg fields
			for iz := 0; iz < nz; iz++ {
				i := I(ix,iy,iz)
				g.rho[i] = wtg.rho[iz]
				g.p[i] = wtg.pres[iz]
				g.drt[i] = wtg.drt[iz]
				g.dent[i] = wtg.dent[iz]
				g.ux[i] = wtg.ux[iz]
				g.uy[i] = wtg.uy[iz]
				g.uz[i] = 0.0
				g.drr[i] = 0.0
				g.keddy[i] = 0.0

				// initial perturbation of entropy field
				zval := g.z[iz]
				g.dent[i] += (rentinit + pointamp)*
					math.Exp(-zval/zscale)
			}

			// initialize cumulative flux fields
			j := I2(ix,iy)
			g.surfentcum[j] = 0.0
			g.surfrtcum[j] = 0.0
			g.raincum[j] = 0.0
		}
	}
}

// *******************************************************************
// compute hydrostatic profile for fixed temperature lapse rate and
// relative humidity -- ignore contribution of moisture to hydrostatic
// balance
func hydrostatic(nz int, temp0, rh, p0, tp, tlapse,
	dz float64) (p, rho, drt, ddryent, dent, dsatent []float64) {

	// make output slices
	p = make([]float64,nz)
	rho = make([]float64,nz)
	drt = make([]float64,nz)
	ddryent = make([]float64,nz)
	dent = make([]float64,nz)
	dsatent = make([]float64,nz)

	// initialize variables at lowest model level
	p[0] = p0
	rho[0] = thermo.Rho(temp0,p0,0.0)
	rsat := thermo.Rsat(temp0,rho[0])
	drt[0] = rh*rsat*rho[0]
	ddryent[0] = thermo.Dent(temp0,rho[0],rh*rsat)*rho[0]
	dent[0] = thermo.Ent(temp0,rho[0],rh*rsat,0.0)*rho[0]
	dsatent[0] = thermo.Sent(temp0,rho[0])*rho[0]
	temp := temp0
	logp := math.Log(p[0])
	ttrop := temp0 - tp*tlapse
	for iz := 1; iz < nz; iz++ {

		// compute half step and full step temperature
		tstar := temp - 0.5*dz*tlapse
		if tstar < ttrop {tstar = ttrop}
		temp = temp - dz*tlapse
		if temp < ttrop {temp = ttrop}

		// compute log(p) and get p
		logp = logp - dz*Gee/(thermo.R*tstar)
		p[iz] = math.Exp(logp)

		// compute rho, rt, sd, s, and ss
		rho[iz] = thermo.Rho(temp,p[iz],0.0)
		rsat = thermo.Rsat(temp,rho[iz])
		drt[iz] = rh*rsat*rho[iz]
		ddryent[iz] = thermo.Dent(temp,rho[iz],rh*rsat)*rho[iz]
		dent[iz] = thermo.Ent(temp,rho[iz],rh*rsat,0.0)*rho[iz]
		dsatent[iz] = thermo.Sent(temp,rho[iz])*rho[iz]
	}
	return
}

// *******************************************************************
// get an initialization profile file
func getinit(nz int, initfile string) (p, rho, drt, ddryent,
	dent, dsatent, ux, uy []float64, vxsys, vysys, sst0 float64) {

	// the init profile is stored in a candis file
	a := gocandis.GetCandis(initfile)

	// check properties of file
	dnames := a.GetDimNames()
	if len(dnames) != 1 {
		fmt.Println("Getinit: input file not 1D")
		os.Exit(1)
	}
	if dnames[0] != "z" {
		fmt.Println("Getinit: dname not z")
		os.Exit(1)
	}
	z := a.GetDimData(dnames[0])
	if len(z) != nz {
		fmt.Println("Getinit: z not the right length")
		os.Exit(1)
	}

	// get SST0, converting to Kelvin
	sst0 = a.GetFieldData("sst")[0] + 273.15

	// get thermodynamic data
	p = a.GetFieldData("pres")
	rho = a.GetFieldData("rho")
	drt = a.GetFieldData("drt")
	ddryent = a.GetFieldData("ddryent")
	dent = a.GetFieldData("dent")
	dsatent = a.GetFieldData("dsatent")

	// get winds
	ux = a.GetFieldData("ux")
	uy = a.GetFieldData("uy")

	// get system propagation velocity
	vxsys = a.GetFieldData("vxsys")[0]
	vysys = a.GetFieldData("vysys")[0]

	// done
	return
}

// *******************************************************************
// *******************************************************************
// *******************************************************************
// *******************************************************************
// *******************************************************************
// scatter global field values to local fields
// taking into account the extra perimeter values
// note that initial global fields are available to all
// processes since these were calculated before
// starting MPI
func scatter(rank, gx, gy, nx, ny, nz int, l *lfields, g *gfields) {
	xloc, yloc := L(rank,gx,gy)
	for ix := 0; ix < mx + 2; ix++ {
		for iy := 0; iy < my + 2; iy++ {
			for iz := 0; iz < nz; iz++ {

				// compute global field indices
				jx := mx*xloc + ix - 1
				if jx < 0 {jx = jx + nx}
				if jx >= nx {jx = jx - nx}
				jy := my*yloc + iy - 1
				if jy < 0 {jy = jy + ny}
				if jy >= ny {jy = jy - ny}

				// local field address
				i := J(ix,iy,iz)

				// global field address
				j := I(jx,jy,iz)

				// transfer data
				l.rho[i] = g.rho[j]
				l.ux[i] = g.ux[j]
				l.uy[i] = g.uy[j]
				l.uz[i] = g.uz[j]
				l.dent[i] = g.dent[j]
				l.p[i] = g.p[j]
				l.drt[i] = g.drt[j]
				l.drr[i] = g.drr[j]
				l.keddy[i] = g.keddy[j]
			}
		}
	}
}

// *******************************************************************
// *******************************************************************
// *******************************************************************
// *******************************************************************
// *******************************************************************
// gathers the appropriate components of the local fields for global fields
// same considerations as for scatter operation
func gather(comm gompi.MPI_Comm, rank, gx, gy, ny, nz int,
	l *lfields, g *gfields) {

	// pack the buffer
	ib0 := 0
	ib0 = lpack(ib0,rank,gx,gy,l.rho[:],l.buff[:])
	ib0 = lpack(ib0,rank,gx,gy,l.ux[:],l.buff[:])
	ib0 = lpack(ib0,rank,gx,gy,l.uy[:],l.buff[:])
	ib0 = lpack(ib0,rank,gx,gy,l.uz[:],l.buff[:])
	ib0 = lpack(ib0,rank,gx,gy,l.dent[:],l.buff[:])
	ib0 = lpack(ib0,rank,gx,gy,l.p[:],l.buff[:])
	ib0 = lpack(ib0,rank,gx,gy,l.drt[:],l.buff[:])
	ib0 = lpack(ib0,rank,gx,gy,l.drr[:],l.buff[:])
	ib0 = lpack(ib0,rank,gx,gy,l.keddy[:],l.buff[:])
	ib0 = lpack(ib0,rank,gx,gy,l.sux[:],l.buff[:])
	ib0 = lpack(ib0,rank,gx,gy,l.suy[:],l.buff[:])
	ib0 = lpack(ib0,rank,gx,gy,l.suz[:],l.buff[:])
	ib0 = lpack(ib0,rank,gx,gy,l.sdent[:],l.buff[:])
	ib0 = lpack(ib0,rank,gx,gy,l.sdrt[:],l.buff[:])
	ib0 = lpack(ib0,rank,gx,gy,l.sdrr[:],l.buff[:])
	ib0 = lpack(ib0,rank,gx,gy,l.sdrad[:],l.buff[:])
	ib0 = lpack(ib0,rank,gx,gy,l.sflux[:],l.buff[:])
	ib0 = lpack(ib0,rank,gx,gy,l.flux[:],l.buff[:])
	ib0 = lpack2(ib0,rank,gx,gy,l.surfvx[:],l.buff[:])
	ib0 = lpack2(ib0,rank,gx,gy,l.surfvy[:],l.buff[:])
	ib0 = lpack2(ib0,rank,gx,gy,l.surfent[:],l.buff[:])
	ib0 = lpack2(ib0,rank,gx,gy,l.surfrt[:],l.buff[:])
	ib0 = lpack2(ib0,rank,gx,gy,l.rain[:],l.buff[:])
	ib0 = lpack2(ib0,rank,gx,gy,l.surfentcum[:],l.buff[:])
	ib0 = lpack2(ib0,rank,gx,gy,l.surfrtcum[:],l.buff[:])
	ib0 = lpack2(ib0,rank,gx,gy,l.raincum[:],l.buff[:])
	ib0 = lpack2(ib0,rank,gx,gy,l.sst[:],l.buff[:])
	ib0 = lpack2(ib0,rank,gx,gy,l.surft[:],l.buff[:])

	// simple copy if rank == 0, otherwise MPI send
	gatherflag := make(chan int,0)
	if rank == 0 {
		blen := len(l.buff)
		for i := 0; i < blen; i++ {
			l.rbuff[i] = l.buff[i]
		}
	} else {
		go func() {
			gompi.Send_float64(l.buff[:],0,comm)
			gatherflag <- 0
		} ()
	}

	// receiver has to loop over all senders
	if rank == 0 {
		for xrank := 0; xrank < gx*gy; xrank++ {
			if xrank != 0 {
				gompi.Recv_float64(l.rbuff[:],xrank,comm)
			}
			
			// receive finished
			ib0 := 0
			ib0 = gunpack(ib0,xrank,gx,gy,ny,nz,g.rho[:],l.rbuff[:])
			ib0 = gunpack(ib0,xrank,gx,gy,ny,nz,g.ux[:],l.rbuff[:])
			ib0 = gunpack(ib0,xrank,gx,gy,ny,nz,g.uy[:],l.rbuff[:])
			ib0 = gunpack(ib0,xrank,gx,gy,ny,nz,g.uz[:],l.rbuff[:])
			ib0 = gunpack(ib0,xrank,gx,gy,ny,nz,g.dent[:],l.rbuff[:])
			ib0 = gunpack(ib0,xrank,gx,gy,ny,nz,g.p[:],l.rbuff[:])
			ib0 = gunpack(ib0,xrank,gx,gy,ny,nz,g.drt[:],l.rbuff[:])
			ib0 = gunpack(ib0,xrank,gx,gy,ny,nz,g.drr[:],l.rbuff[:])
			ib0 = gunpack(ib0,xrank,gx,gy,ny,nz,g.keddy[:],l.rbuff[:])
			ib0 = gunpack(ib0,xrank,gx,gy,ny,nz,g.sux[:],l.rbuff[:])
			ib0 = gunpack(ib0,xrank,gx,gy,ny,nz,g.suy[:],l.rbuff[:])
			ib0 = gunpack(ib0,xrank,gx,gy,ny,nz,g.suz[:],l.rbuff[:])
			ib0 = gunpack(ib0,xrank,gx,gy,ny,nz,g.sdent[:],l.rbuff[:])
			ib0 = gunpack(ib0,xrank,gx,gy,ny,nz,g.sdrt[:],l.rbuff[:])
			ib0 = gunpack(ib0,xrank,gx,gy,ny,nz,g.sdrr[:],l.rbuff[:])
			ib0 = gunpack(ib0,xrank,gx,gy,ny,nz,g.sdrad[:],l.rbuff[:])
			ib0 = gunpack(ib0,xrank,gx,gy,ny,nz,g.sflux[:],l.rbuff[:])
			ib0 = gunpack(ib0,xrank,gx,gy,ny,nz,g.flux[:],l.rbuff[:])
			ib0 = gunpack2(ib0,xrank,gx,gy,ny,g.surfvx[:],l.rbuff[:])
			ib0 = gunpack2(ib0,xrank,gx,gy,ny,g.surfvy[:],l.rbuff[:])
			ib0 = gunpack2(ib0,xrank,gx,gy,ny,g.surfent[:],l.rbuff[:])
			ib0 = gunpack2(ib0,xrank,gx,gy,ny,g.surfrt[:],l.rbuff[:])
			ib0 = gunpack2(ib0,xrank,gx,gy,ny,g.rain[:],l.rbuff[:])
			ib0 = gunpack2(ib0,xrank,gx,gy,ny,g.surfentcum[:],l.rbuff[:])
			ib0 = gunpack2(ib0,xrank,gx,gy,ny,g.surfrtcum[:],l.rbuff[:])
			ib0 = gunpack2(ib0,xrank,gx,gy,ny,g.raincum[:],l.rbuff[:])
			ib0 = gunpack2(ib0,xrank,gx,gy,ny,g.sst[:],l.rbuff[:])
			ib0 = gunpack2(ib0,xrank,gx,gy,ny,g.surft[:],l.rbuff[:])
		}
	} else {

		// finished with sends
		<- gatherflag
	}
}

// *******************************************************************
// pack the part of the local field needed for the global field
// rank into a buffer
func lpack(ib0 int, rank, gx, gy int, lfield, buff []float64) (newib0 int) {

	// make transfer
	newib0 = ib0
	for ix := 0; ix < mx; ix++ {
		for iy := 0; iy < my; iy++ {
			for iz := 0; iz < mz; iz++ {
				j := J(ix + 1,iy + 1,iz)
				buff[newib0] = lfield[j]
				newib0 = newib0 + 1
			}
		}
	}
	return
}

// *******************************************************************
// pack the part of the local field needed for the global field
// rank into a buffer
func lpack2(ib0 int, rank, gx, gy int, lfield, buff []float64) (newib0 int) {

	// make transfer
	newib0 = ib0
	for ix := 0; ix < mx; ix++ {
		for iy := 0; iy < my; iy++ {
			j := J2(ix + 1,iy + 1)
			buff[newib0] = lfield[j]
			newib0 = newib0 + 1
		}
	}
	return
}

// *******************************************************************
// unpack the buffer into the corresponding local fields
func gunpack(ib0 int, rank, gx, gy, ny, nz int,
	gfield,	buff []float64) (newib0 int) {

	// get global field starting point
	xloc, yloc := L(rank,gx,gy)

	// make transfer
	newib0 = ib0
	for ix := 0; ix < mx; ix++ {
		for iy := 0; iy < my; iy++ {
			for iz := 0; iz < mz; iz++ {
				j := I(ix + mx*xloc,iy + my*yloc,iz)
				gfield[j] = buff[newib0]
				newib0 = newib0 + 1
			}
		}
	}
	return
}

// *******************************************************************
// unpack the buffer into the corresponding local fields
func gunpack2(ib0 int, rank, gx, gy, ny int,
	gfield,	buff []float64) (newib0 int) {

	// get global field starting point
	xloc, yloc := L(rank,gx,gy)

	// make transfer
	newib0 = ib0
	for ix := 0; ix < mx; ix++ {
		for iy := 0; iy < my; iy++ {
			j := I2(ix + mx*xloc,iy + my*yloc)
			gfield[j] = buff[newib0]
			newib0 = newib0 + 1
		}
	}
	return
}

// *******************************************************************
// *******************************************************************
// *******************************************************************
// *******************************************************************
// *******************************************************************
// do a time step incorporating both fine and coarse loops
func timestep(comm gompi.MPI_Comm, rank, nx, ny, nz, it int, l *lfields,
	g *gfields, wtg *wtgfields, paramlist map[string]string) {

	// get parameters
	lt := params.Iparam(paramlist,"lt")
	mt := params.Iparam(paramlist,"mt")
	dx := params.Fparam(paramlist,"dx")
	dy := params.Fparam(paramlist,"dy")
	dz := params.Fparam(paramlist,"dz")
	dt := params.Fparam(paramlist,"dt")
	f0 := params.Fparam(paramlist,"f0")

	// half timestep
	dh := dt/2.0

	// coarse scale time loop
	for jt := 0; jt < mt; jt++ {
		
		// compute coarse scale time
		time := float64(it*mt*lt + jt*lt)*dt

		// fine scale time loop
		for kt := 0; kt < lt; kt++ {

			// compute unstaggered grid fluxes
			for i := 0; i < lsize; i++ {
				rho := l.rho[i]
				vx := l.ux[i]/rho
				vy := l.uy[i]/rho
				vz := l.uz[i]/rho
				l.fxx[i] = vx*l.ux[i]
				l.fyy[i] = vy*l.uy[i]
				l.fzz[i] = vz*l.uz[i]
				l.fxy[i] = vx*l.uy[i]
				l.fyz[i] = vy*l.uz[i]
				l.fzx[i] = vz*l.ux[i]
				l.fxent[i] = vx*l.dent[i]
				l.fyent[i] = vy*l.dent[i]
				l.fzent[i] = vz*l.dent[i]
				l.fxrt[i] = vx*l.drt[i]
				l.fyrt[i] = vy*l.drt[i]
				l.fzrt[i] = vz*l.drt[i]
				l.fxrr[i] = vx*l.drr[i]
				l.fyrr[i] = vy*l.drr[i]
				l.fzrr[i] = vz*l.drr[i]
			}
			
			// loop over staggered grid horizontally internal
			// to unstaggered grid
			// making a fast half step calculation
			for ix := 0; ix < mx + 1; ix++ {
				for iy := 0; iy < my + 1; iy++ {
					
					// calculate staggered intermediate values
					// excluding top and bottom layers
					for iz := 1; iz < mz; iz++ {
						halfstep(ix,iy,iz,dx,dy,dz,
							dh,f0,l)
					}
					
					// extrapolate to upper and lower
					// boundaries on staggered grid
					ulbound(ix,iy,l)
				}
			}

			// compute staggered grid fluxes
			for i := 0; i < ilsize; i++ {
				rho := l.irho[i]
				vx := l.iux[i]/rho
				vy := l.iuy[i]/rho
				vz := l.iuz[i]/rho
				l.fxx[i] = vx*l.iux[i]
				l.fyy[i] = vy*l.iuy[i]
				l.fzz[i] = vz*l.iuz[i]
				l.fxy[i] = vx*l.iuy[i]
				l.fyz[i] = vy*l.iuz[i]
				l.fzx[i] = vz*l.iux[i]
				l.fxent[i] = vx*l.ident[i]
				l.fyent[i] = vy*l.ident[i]
				l.fzent[i] = vz*l.ident[i]
				l.fxrt[i] = vx*l.idrt[i]
				l.fyrt[i] = vy*l.idrt[i]
				l.fzrt[i] = vz*l.idrt[i]
				l.fxrr[i] = vx*l.idrr[i]
				l.fyrr[i] = vy*l.idrr[i]
				l.fzrr[i] = vz*l.idrr[i]
			}

			// set up compensation for negative rt
			l.drtnegative = 0.0
			l.drttotal = 0.0
			
			// do full step on one cell in from periphery
			// to prepare for transfers to neighbors on
			// unstaggered grid
			// do z iteration from top down so that iz == 0
			// interpolation is based on current data
			for iz := mz - 1; iz >= 0; iz-- {
				
				// E/W walls -- get corners
				for iy := 1; iy < my + 1; iy++ {
					fullstep(1,iy,iz,dx,dy,dz,dt,f0,l,wtg)

					// if mx == 1 this is a repeat operation
					if mx > 1 {
						fullstep(mx,iy,iz,dx,dy,dz,dt,
							f0,l,wtg)
					}
				}
				
				// N/S walls -- don't repeat corners
				for ix := 2; ix < mx; ix++ {
					fullstep(ix,1,iz,dx,dy,dz,dt,f0,l,wtg)

					// if my == 1 this is a repeat operation
					if my > 1 {
						fullstep(ix,my,iz,dx,dy,dz,dt,
							f0,l,wtg)
					}
				}
			}
			
			// do interior calculations in a gofunction so it
			// doesn't block the edge transfers
			intcalcs := make(chan int,0)
			go func() {
				for ix := 2; ix < mx; ix++ {
					for iy := 2; iy < my; iy++ {
						for iz := mz - 1; iz >= 0; iz-- {
							fullstep(ix,iy,iz,dx,dy,dz,dt,
								f0,l,wtg)
						}
					}
				}

				// indicate that we are finished
				intcalcs <- 0
			} ()

			// do edge transfers
			edgetrans(comm,rank,gx,gy,l)

			// wait until interior calculations are done
			<- intcalcs
			
			// maintain the integrated rt in the fine timestep
			compensation := l.drttotal/(l.drttotal + l.drtnegative)
			for ix := 0; ix < lx; ix++ {
				for iy := 0; iy < ly; iy++ {
					for iz := 0; iz < lz; iz++ {
						j := J(ix,iy,iz)
						l.drt[j] = compensation*l.drt[j]
					}
				}
			}
		}
			
		// do local coarse time step stuff
		coarse(comm,rank,gx,gy,lt,dx,dy,dz,dt,time,l,wtg,paramlist)

		// gather wtg-related stuff
		gatherwtg(comm,nz,rank,gx,gy,l,wtg)
		
		// compute source terms for wtg-related effects
		if rank == 0 {
			wtgcalc(nx,ny,nz,g,wtg,paramlist)
		}

		// scatter wtg-related stuff
		scatterwtg(comm,nz,rank,gx,gy,wtg)
	}
}

// *******************************************************************
// do the edge transfers in fine loop
func edgetrans(comm gompi.MPI_Comm, rank, gx, gy int, l *lfields) {
	
	// find out where we are
	xloc, yloc := L(rank,gx,gy)
	
	// find ranks of nearest neighbors
	nrank := M(xloc,yloc + 1,gx,gy)
	srank := M(xloc,yloc - 1,gx,gy)
	erank := M(xloc + 1,yloc,gx,gy)
	wrank := M(xloc - 1,yloc,gx,gy)
	
	// do edge transfers one at a time, n/s first then e/w
	dofilln(l)
	dotrans(l.sbuffn[:],l.rbuffs[:],yloc,rank,nrank,srank,comm)
	doextrs(l)
	dofills(l)
	dotrans(l.sbuffs[:],l.rbuffn[:],yloc,rank,srank,nrank,comm)
	doextrn(l)
	dofille(l)
	dotrans(l.sbuffe[:],l.rbuffw[:],xloc,rank,erank,wrank,comm)
	doextrw(l)
	dofillw(l)
	dotrans(l.sbuffw[:],l.rbuffe[:],xloc,rank,wrank,erank,comm)
	doextre(l)
}

// *******************************************************************
// do a transfer
func dotrans(sbuff, rbuff []float64, loc, rank, srank, rrank int,
	comm gompi.MPI_Comm) {

	// if all the same ranks, just copy, if not, MPI transfer
	if srank == rank && rrank == rank {
		nelements := len(sbuff)
		for i := 0; i < nelements; i++ {
			rbuff[i] = sbuff[i]
		}
	} else {

		// odd to even
		hloc := loc/2
		if loc - 2*hloc == 1 {
			gompi.Send_float64(sbuff,srank,comm)
		} else {
			gompi.Recv_float64(rbuff,rrank,comm)
		}

		// even to odd
		if loc - 2*hloc == 0 {
			gompi.Send_float64(sbuff,srank,comm)
		} else {
			gompi.Recv_float64(rbuff,rrank,comm)
		}
	}
}

// *******************************************************************
// fill the send buffers -- n/s skip corners, e/w include corners
func dofilln(l *lfields) {
	j := 0
	for iz := 0; iz < mz; iz++ {
		for ix := 1; ix < mx + 1; ix++ {
			i := J(ix,my,iz)
			l.sbuffn[j] = l.rho[i]; j++
			l.sbuffn[j] = l.ux[i]; j++
			l.sbuffn[j] = l.uy[i]; j++
			l.sbuffn[j] = l.uz[i]; j++
			l.sbuffn[j] = l.dent[i]; j++
			l.sbuffn[j] = l.p[i]; j++
			l.sbuffn[j] = l.drt[i]; j++
			l.sbuffn[j] = l.drr[i]; j++
			l.sbuffn[j] = l.sux[i]; j++
			l.sbuffn[j] = l.suy[i]; j++
			l.sbuffn[j] = l.suz[i]; j++
			l.sbuffn[j] = l.sdent[i]; j++
			l.sbuffn[j] = l.sdrt[i]; j++
			l.sbuffn[j] = l.sdrr[i]; j++
		}
	}
}
func dofills(l *lfields) {
	j := 0
	for iz := 0; iz < mz; iz++ {
		for ix := 1; ix < mx + 1; ix++ {
			i := J(ix,1,iz)
			l.sbuffs[j] = l.rho[i]; j++
			l.sbuffs[j] = l.ux[i]; j++
			l.sbuffs[j] = l.uy[i]; j++
			l.sbuffs[j] = l.uz[i]; j++
			l.sbuffs[j] = l.dent[i]; j++
			l.sbuffs[j] = l.p[i]; j++
			l.sbuffs[j] = l.drt[i]; j++
			l.sbuffs[j] = l.drr[i]; j++
			l.sbuffs[j] = l.sux[i]; j++
			l.sbuffs[j] = l.suy[i]; j++
			l.sbuffs[j] = l.suz[i]; j++
			l.sbuffs[j] = l.sdent[i]; j++
			l.sbuffs[j] = l.sdrt[i]; j++
			l.sbuffs[j] = l.sdrr[i]; j++
		}
	}
}
func dofille(l *lfields) {
	j := 0
	for iz := 0; iz < mz; iz++ {
		for iy := 0; iy < my + 2; iy++ {
			i := J(mx,iy,iz)
			l.sbuffe[j] = l.rho[i]; j++
			l.sbuffe[j] = l.ux[i]; j++
			l.sbuffe[j] = l.uy[i]; j++
			l.sbuffe[j] = l.uz[i]; j++
			l.sbuffe[j] = l.dent[i]; j++
			l.sbuffe[j] = l.p[i]; j++
			l.sbuffe[j] = l.drt[i]; j++
			l.sbuffe[j] = l.drr[i]; j++
			l.sbuffe[j] = l.sux[i]; j++
			l.sbuffe[j] = l.suy[i]; j++
			l.sbuffe[j] = l.suz[i]; j++
			l.sbuffe[j] = l.sdent[i]; j++
			l.sbuffe[j] = l.sdrt[i]; j++
			l.sbuffe[j] = l.sdrr[i]; j++
		}
	}
}
func dofillw(l *lfields) {
	j := 0
	for iz := 0; iz < mz; iz++ {
		for iy := 0; iy < my + 2; iy++ {
			i := J(1,iy,iz)
			l.sbuffw[j] = l.rho[i]; j++
			l.sbuffw[j] = l.ux[i]; j++
			l.sbuffw[j] = l.uy[i]; j++
			l.sbuffw[j] = l.uz[i]; j++
			l.sbuffw[j] = l.dent[i]; j++
			l.sbuffw[j] = l.p[i]; j++
			l.sbuffw[j] = l.drt[i]; j++
			l.sbuffw[j] = l.drr[i]; j++
			l.sbuffw[j] = l.sux[i]; j++
			l.sbuffw[j] = l.suy[i]; j++
			l.sbuffw[j] = l.suz[i]; j++
			l.sbuffw[j] = l.sdent[i]; j++
			l.sbuffw[j] = l.sdrt[i]; j++
			l.sbuffw[j] = l.sdrr[i]; j++
		}
	}
}

// *******************************************************************
// empty the receive buffers -- n/s skip corners, e/w include corners
func doextrn(l *lfields) {
	j := 0
	for iz := 0; iz < mz; iz++ {
		for ix := 1; ix < mx + 1; ix++ {
			i := J(ix,my + 1,iz)
			l.rho[i] = l.rbuffn[j]; j++
			l.ux[i] = l.rbuffn[j]; j++
			l.uy[i] = l.rbuffn[j]; j++
			l.uz[i] = l.rbuffn[j]; j++
			l.dent[i] = l.rbuffn[j]; j++
			l.p[i] = l.rbuffn[j]; j++
			l.drt[i] = l.rbuffn[j]; j++
			l.drr[i] = l.rbuffn[j]; j++
			l.sux[i] = l.rbuffn[j]; j++
			l.suy[i] = l.rbuffn[j]; j++
			l.suz[i] = l.rbuffn[j]; j++
			l.sdent[i] = l.rbuffn[j]; j++
			l.sdrt[i] = l.rbuffn[j]; j++
			l.sdrr[i] = l.rbuffn[j]; j++
		}
	}
}
func doextrs(l *lfields) {
	j := 0
	for iz := 0; iz < mz; iz++ {
		for ix := 1; ix < mx + 1; ix++ {
			i := J(ix,0,iz)
			l.rho[i] = l.rbuffs[j]; j++
			l.ux[i] = l.rbuffs[j]; j++
			l.uy[i] = l.rbuffs[j]; j++
			l.uz[i] = l.rbuffs[j]; j++
			l.dent[i] = l.rbuffs[j]; j++
			l.p[i] = l.rbuffs[j]; j++
			l.drt[i] = l.rbuffs[j]; j++
			l.drr[i] = l.rbuffs[j]; j++
			l.sux[i] = l.rbuffs[j]; j++
			l.suy[i] = l.rbuffs[j]; j++
			l.suz[i] = l.rbuffs[j]; j++
			l.sdent[i] = l.rbuffs[j]; j++
			l.sdrt[i] = l.rbuffs[j]; j++
			l.sdrr[i] = l.rbuffs[j]; j++
		}
	}
}
func doextre(l *lfields) {
	j := 0
	for iz := 0; iz < mz; iz++ {
		for iy := 0; iy < my + 2; iy++ {
			i := J(mx + 1,iy,iz)
			l.rho[i] = l.rbuffe[j]; j++
			l.ux[i] = l.rbuffe[j]; j++
			l.uy[i] = l.rbuffe[j]; j++
			l.uz[i] = l.rbuffe[j]; j++
			l.dent[i] = l.rbuffe[j]; j++
			l.p[i] = l.rbuffe[j]; j++
			l.drt[i] = l.rbuffe[j]; j++
			l.drr[i] = l.rbuffe[j]; j++
			l.sux[i] = l.rbuffe[j]; j++
			l.suy[i] = l.rbuffe[j]; j++
			l.suz[i] = l.rbuffe[j]; j++
			l.sdent[i] = l.rbuffe[j]; j++
			l.sdrt[i] = l.rbuffe[j]; j++
			l.sdrr[i] = l.rbuffe[j]; j++
		}
	}
}
func doextrw(l *lfields) {
	j := 0
	for iz := 0; iz < mz; iz++ {
		for iy := 0; iy < my + 2; iy++ {
			i := J(0,iy,iz)
			l.rho[i] = l.rbuffw[j]; j++
			l.ux[i] = l.rbuffw[j]; j++
			l.uy[i] = l.rbuffw[j]; j++
			l.uz[i] = l.rbuffw[j]; j++
			l.dent[i] = l.rbuffw[j]; j++
			l.p[i] = l.rbuffw[j]; j++
			l.drt[i] = l.rbuffw[j]; j++
			l.drr[i] = l.rbuffw[j]; j++
			l.sux[i] = l.rbuffw[j]; j++
			l.suy[i] = l.rbuffw[j]; j++
			l.suz[i] = l.rbuffw[j]; j++
			l.sdent[i] = l.rbuffw[j]; j++
			l.sdrt[i] = l.rbuffw[j]; j++
			l.sdrr[i] = l.rbuffw[j]; j++
		}
	}
}

// *******************************************************************
// do the half step calculation for a specified grid cell
func halfstep(ix, iy, iz int, dx, dy, dz, dh, f0 float64, l *lfields) {

	// index for receiving intermediate fields
	j := iJ(ix,iy,iz)
	
	// density
	l.irho[j] = Ctr(ix,iy,iz,l.rho[:]) -
		(Cxd(ix,iy,iz,dx,l.ux[:]) +
		Cyd(ix,iy,iz,dy,l.uy[:]) +
		Czd(ix,iy,iz,dz,l.uz[:]))*dh

	// x momentum density
	l.iux[j] = Ctr(ix,iy,iz,l.ux[:]) -
		(Cxd(ix,iy,iz,dx,l.p[:]) +
		Cxd(ix,iy,iz,dx,l.fxx[:]) +
		Cyd(ix,iy,iz,dy,l.fxy[:]) +
		Czd(ix,iy,iz,dz,l.fzx[:]))*dh +
		f0*Ctr(ix,iy,iz,l.uy[:])*dh

	// y momentum density
	l.iuy[j] = Ctr(ix,iy,iz,l.uy[:]) -
		(Cyd(ix,iy,iz,dy,l.p[:]) +
		Cxd(ix,iy,iz,dx,l.fxy[:]) +
		Cyd(ix,iy,iz,dy,l.fyy[:]) +
		Czd(ix,iy,iz,dz,l.fyz[:]))*dh -
		f0*Ctr(ix,iy,iz,l.ux[:])*dh

	// z momentum density
	l.iuz[j] = Ctr(ix,iy,iz,l.uz[:]) -
		(Czd(ix,iy,iz,dz,l.p[:]) +
		Cxd(ix,iy,iz,dx,l.fzx[:]) +
		Cyd(ix,iy,iz,dy,l.fyz[:]) +
		Czd(ix,iy,iz,dz,l.fzz[:]))*dh -
		Gee*(Ctr(ix,iy,iz,l.rho[:]) +
		Ctr(ix,iy,iz,l.drt[:]) +
		Ctr(ix,iy,iz,l.drr[:]))*dh

	// entropy density
	l.ident[j] = Ctr(ix,iy,iz,l.dent[:]) -
		(Cxd(ix,iy,iz,dx,l.fxent[:]) +
		Cyd(ix,iy,iz,dy,l.fyent[:]) +
		Czd(ix,iy,iz,dz,l.fzent[:]))*dh

	// total cloud water density
	l.idrt[j] = Ctr(ix,iy,iz,l.drt[:]) -
		(Cxd(ix,iy,iz,dx,l.fxrt[:]) +
		Cyd(ix,iy,iz,dy,l.fyrt[:]) +
		Czd(ix,iy,iz,dz,l.fzrt[:]))*dh

	// precipitation density
	l.idrr[j] = Ctr(ix,iy,iz,l.drr[:]) -
		(Cxd(ix,iy,iz,dx,l.fxrr[:]) +
		Cyd(ix,iy,iz,dy,l.fyrr[:]) +
		Czd(ix,iy,iz,dz,l.fzrr[:]))*dh

	// diagnose pressure from density, cloud water
	// mixing ratio, precip mixing ratio, and total moist entropy
	l.ip[j] = thermo.Pres2(l.irho[j],l.idrt[j]/l.irho[j],
		l.idrr[j]/l.irho[j],l.ident[j]/l.irho[j])
}

// *******************************************************************
// extension conditions on upper and lower boundaries for staggered grid
func ulbound(ix, iy int, l *lfields) {

	// lower boundary -- thermo zeroth order, wind zeroth
	i := iJ(ix,iy,1)
	j := iJ(ix,iy,0)
	k := iJ(ix,iy,2)
	l.irho[j] = 2.0*l.irho[i] - l.irho[k]
	l.iux[j] = l.iux[i]
	l.iuy[j] = l.iuy[i]
	l.iuz[j] = 0.0
	l.ident[j] = 2.0*l.ident[i] - l.ident[k]
	l.idrt[j] = 2.0*l.idrt[i] - l.idrt[k]
	l.idrr[j] = 2.0*l.idrr[i] - l.idrr[k]

	// don't allow rain mixing ratio to become negative
	if l.idrr[j] < 0.0 {l.idrr[j] = 0.0}
	l.ip[j] = thermo.Pres2(l.irho[j],l.idrt[j]/l.irho[j],
		l.idrr[j]/l.irho[j],l.ident[j]/l.irho[j])

	// upper boundary
	i = iJ(ix,iy,mz - 1)
	j = iJ(ix,iy,mz)
	k = iJ(ix,iy,mz - 2)
	l.irho[j] = 2.0*l.irho[i] - l.irho[k]
	l.iux[j] = l.iux[i]
	l.iuy[j] = l.iuy[i]
	l.iuz[j] = 0.0
	l.ident[j] = 2.0*l.ident[i] - l.ident[k]
	l.idrt[j] = 2.0*l.idrt[i] - l.idrt[k]
	l.idrr[j] = 2.0*l.idrr[i] - l.idrr[k]

	// don't allow rain mixing ratio to become negative
	if l.idrr[j] < 0.0 {l.idrr[j] = 0.0}
	l.ip[j] = thermo.Pres2(l.irho[j],l.idrt[j]/l.irho[j],
		l.idrr[j]/l.irho[j],l.ident[j]/l.irho[j])
}

	// *******************************************************************
func fullstep(ix, iy, iz int, dx, dy, dz, dt, f0 float64, l *lfields,
	wtg *wtgfields) {

	// index for receiving final fields
	i := J(ix,iy,iz)
	j := J(ix,iy,iz + 1)
	k := J(ix,iy,iz - 1)

	// density
	l.rho[i] = l.rho[i] -
		(Fxd(ix,iy,iz,dx,l.iux[:]) +
		Fyd(ix,iy,iz,dy,l.iuy[:]) +
		Fzd(ix,iy,iz,dz,l.iuz[:]))*dt

	// x momentum density
	l.ux[i] = l.ux[i] -
		((Fxd(ix,iy,iz,dx,l.ip[:]) +
		Fxd(ix,iy,iz,dx,l.fxx[:]) +
		Fyd(ix,iy,iz,dy,l.fxy[:]) +
		Fzd(ix,iy,iz,dz,l.fzx[:])) -
		f0*Ftr(ix,iy,iz,l.iuy[:]) +
		wtg.mu[iz]*(l.ux[i] - wtg.ux[iz]))*dt

	// y momentum density
	l.uy[i] = l.uy[i] -
		((Fyd(ix,iy,iz,dy,l.ip[:]) +
		Fxd(ix,iy,iz,dx,l.fxy[:]) +
		Fyd(ix,iy,iz,dy,l.fyy[:]) +
		Fzd(ix,iy,iz,dz,l.fyz[:])) +
		f0*Ftr(ix,iy,iz,l.iux[:]) +
		wtg.mu[iz]*(l.uy[i] - wtg.uy[iz]))*dt

	// z momentum density -- special treatment at boundaries
	// note that iz = 0 case gets current uz at iz = 1 due
	// to downward iteration when fullstep is called --
	// top case gets previous data, which is less important
	if iz == 0 {
		l.uz[i] = l.uz[j]/3.0
	} else if iz == mz - 1 {
		l.uz[i] = l.uz[k]/3.0
	} else {
		l.uz[i] = l.uz[i] -
			((Fzd(ix,iy,iz,dz,l.ip[:]) +
			Fxd(ix,iy,iz,dx,l.fzx[:]) +
			Fyd(ix,iy,iz,dy,l.fyz[:]) +
			Fzd(ix,iy,iz,dz,l.fzz[:])) +
			Gee*(Ftr(ix,iy,iz,l.irho[:]) +
			Ftr(ix,iy,iz,l.idrt[:]) +
			Ftr(ix,iy,iz,l.idrr[:])) +
			wtg.mu[iz]*l.uz[i])*dt
	}

	// entropy density
	l.dent[i] = l.dent[i] -
		(Fxd(ix,iy,iz,dx,l.fxent[:]) +
		Fyd(ix,iy,iz,dy,l.fyent[:]) +
		Fzd(ix,iy,iz,dz,l.fzent[:]))*dt

	// total cloud water density
	l.drt[i] = l.drt[i] -
		(Fxd(ix,iy,iz,dx,l.fxrt[:]) +
		Fyd(ix,iy,iz,dy,l.fyrt[:]) +
		Fzd(ix,iy,iz,dz,l.fzrt[:]))*dt

	// precipitation density
	l.drr[i] = l.drr[i] -
		(Fxd(ix,iy,iz,dx,l.fxrr[:]) +
		Fyd(ix,iy,iz,dy,l.fyrr[:]) +
		Fzd(ix,iy,iz,dz,l.fzrr[:]))*dt

	// this is a hack -- don't let drr or drt become negative
	// integrated drt (cloud water mass) is maintained
	// at a constant value after the fullstep completes
	// if drr is negative, the deficit is transferred to drt first
	l.drttotal += l.drt[i]
	if l.drr[i] < 0.0 {
		l.drt[i] += l.drr[i]
		l.drr[i] = 0.0
	}
	if l.drt[i] < 0.0 {
		l.drtnegative -= l.drt[i]
		l.drt[i] = 0.0
	}

	// diagnose pressure from density, total cloud water mixing
	// ratio, and moist entropy
	l.p[i] = thermo.Pres2(l.rho[i],l.drt[i]/l.rho[i],l.drr[i]/l.rho[i],
		l.dent[i]/l.rho[i])
}		

// *******************************************************************
// compute coarse time step stuff
func coarse(comm gompi.MPI_Comm, rank, gx, gy, lt int, dx, dy, dz, dt, time float64,
	l *lfields, wtg *wtgfields, paramlist map[string]string) {

	// coarse timestep
	dtcoarse := dt*float64(lt)

	// get parameters particular to coarse time step
	eddy := params.Fparam(paramlist,"eddy")
	lambdah := params.Fparam(paramlist,"lambdah")
	lambdaz := params.Fparam(paramlist,"lambdaz")
	lambdaw := params.Fparam(paramlist,"lambdaw")
	delrh := params.Fparam(paramlist,"delrh")
	crain := params.Fparam(paramlist,"crain")
	csnow := params.Fparam(paramlist,"csnow")
	cevap := params.Fparam(paramlist,"cevap")
	wtrain := params.Fparam(paramlist,"wtrain")
	wtsnow := params.Fparam(paramlist,"wtsnow")
	cdrag := params.Fparam(paramlist,"cdrag")
	W := params.Fparam(paramlist,"W")
	tp := params.Fparam(paramlist,"tp")
	radskp := params.Iparam(paramlist,"radskp")
	sun := params.Iparam(paramlist,"sun")
	cld := params.Fparam(paramlist,"cld")
	solarabs := params.Fparam(paramlist,"solarabs")
	solarwv := params.Fparam(paramlist,"solarwv")
	radbrk := params.Fparam(paramlist,"radbrk")
	radheat := params.Fparam(paramlist,"radheat")
	ranent := params.Fparam(paramlist,"ranent")
	zscale := params.Fparam(paramlist,"zscale")
	bowen := params.Fparam(paramlist,"bowen")

	// numerical eddy mixing coefficients
	Kv := lambdaz*dz*dz/dt
	Kw := lambdaw*dz*dz/dt
	
	// divide unstaggered fields by density to get mixing ratios
	for i := 0; i < lsize; i++ {
		l.vx[i] = l.ux[i]/l.rho[i]
		l.vy[i] = l.uy[i]/l.rho[i]
		l.vz[i] = l.uz[i]/l.rho[i]
		l.ent[i] = l.dent[i]/l.rho[i]
		l.rt[i] = l.drt[i]/l.rho[i]
		l.rr[i] = l.drr[i]/l.rho[i]
		
		// we also need the dry entropy
		temp := thermo.Temp(l.rho[i],l.p[i],l.rt[i])
		l.dryent[i] = thermo.Dent(temp,l.rho[i],l.rt[i])
		l.ddryent[i] = l.rho[i]*l.dryent[i]
	}

	// compute surface stuff on staggered grid
	delt := 0.5  // (K)
	for ix := 0; ix < ilx; ix++ {
		for iy := 0; iy < ily; iy++ {
			j2 := iJ2(ix,iy)

			// sea surface temperature
			l.isst[j2] = l.sst0

			// surface fluxes from staggered grid
			// add in domain propagation to get vx,vy rel to earth
			j := iJ(ix,iy,0)
			irho := l.irho[j]
			ivx := (l.iux[j]/irho + wtg.vxsys)
			ivy := (l.iuy[j]/irho + wtg.vysys)
			ient := l.ident[j]/irho
			irt := l.idrt[j]/irho
			ip := l.ip[j]

			// surface fluxes only when SST exceeds air temp
			l.isurft[j2] = thermo.Temp(irho,ip,irt)
			throttle := (l.isst[j2] - l.isurft[j2])/delt
			if throttle > 1.0 {throttle = 1.0}
			if throttle < 0.0 {throttle = 0.0}

			// ueff includes a gustiness factor
			ueff := math.Pow(ivx*ivx + ivy*ivy + W*W,0.5)
			coef := throttle*irho*cdrag*ueff

			// get saturated surface entropy
			isent := thermo.Sent(l.isst[j2],irho)
			isrt := thermo.Rsat(l.isst[j2],irho)

			// do fluxes
			l.isurfvx[j2] = -coef*ivx
			l.isurfvy[j2] = -coef*ivy
			l.isurfent[j2] = coef*(isent - ient)

			// the bowen factor = 1 over the ocean
			// it is between 0 and 1 over the land
			// depending on how moist the surface is
			l.isurfrt[j2] = bowen*coef*(isrt - irt)
		}
	}

	// compute the eddy fluxes on staggered grid
	// exclude upper and lower boundaries, which
	// are considered separately below
	for ix := 0; ix <  ilx; ix++ {
		for iy := 0; iy < ily; iy++ {

			// compute profile of the flux of precip entropy
			// on a staggered vertical grid
			for iz := 0; iz < ilz; iz++ {
				i := iJ(ix,iy,iz)

				// compute the flux of precip entropy
				temp := thermo.Temp(l.irho[i],l.ip[i],
					l.drt[i]/l.rho[i])
				wt := thermo.Ftrans(wtrain,wtsnow,temp)
				l.entdrrflux[iz] = -wt*
					thermo.Entrr(temp,l.idrr[i])
			}

			// compute contributions to slow tendencies
			for iz := mz - 1; iz > 0; iz-- {

				// first, the wind gradients
				dxx := Cxd(ix,iy,iz,dx,l.vx[:])
				dyy := Cyd(ix,iy,iz,dy,l.vy[:])
				dzz := Czd(ix,iy,iz,dz,l.vz[:])
				dxy := 0.5*(Cxd(ix,iy,iz,dx,l.vy[:]) +
					Cyd(ix,iy,iz,dy,l.vx[:]))
				dyz := 0.5*(Cyd(ix,iy,iz,dy,l.vz[:]) +
					Czd(ix,iy,iz,dz,l.vy[:]))
				dzx := 0.5*(Czd(ix,iy,iz,dz,l.vx[:]) +
					Cxd(ix,iy,iz,dx,l.vz[:]))
				dsq := dxx*dxx + dyy*dyy + dzz*dzz +
					2.0*(dxy*dxy + dyz*dyz + dzx*dzx)

				// next, the thermodynamic gradients
				dentx := Cxd(ix,iy,iz,dx,l.ent[:])
				denty := Cyd(ix,iy,iz,dy,l.ent[:])
				dentz := Czd(ix,iy,iz,dz,l.ent[:])
				drtx := Cxd(ix,iy,iz,dx,l.rt[:])
				drty := Cyd(ix,iy,iz,dy,l.rt[:])
				drtz := Czd(ix,iy,iz,dz,l.rt[:])

				drrx := Cxd(ix,iy,iz,dx,l.rr[:])
				drry := Cyd(ix,iy,iz,dy,l.rr[:])
				drrz := Czd(ix,iy,iz,dz,l.rr[:])

				// next, the moist/dry Brunt frequency
				i := iJ(ix,iy,iz)
				rho := l.irho[i]
				rt := l.drt[i]/rho
				temp := thermo.Temp(rho,l.ip[i],rt)
				rs := thermo.Rsat(temp,rho)
				H := l.idrt[i]/(rho*rs)  // relative hum

				// gradual transition from saturated to
				// unsaturated
				nu := 0.5*((H - 1.0)/delrh + 1.0)
				if nu < 0.0 {nu = 0.0}
				if nu > 1.0 {nu = 1.0}

				// dry and moist vertical entropy gradients
				dentdz := Czd(ix,iy,iz,dz,l.ent[:])
				ddryentdz := Czd(ix,iy,iz,dz,l.dryent[:])
 
				// Brunt frequency
				nsq := (Gee/thermo.Cp)*(nu*dentdz +
					(1.0 - nu)*ddryentdz)

				// compute the physical eddy mixing
				// coefficient
				E := dsq - 2.0*nsq
				if E < 0.0 {E = 0.0}
				E = math.Pow(E,0.5)
				K := eddy*rho*E*dz*dz
				l.ikeddy[i] = K

				// this is used only for precipitation
				Ktotz := K + rho*Kv

				// the magnitude of Kw mixing decreases
				// with height, zeroing at tropopause
				hfactor := 1.0 - float64(iz)*dz/tp
				if hfactor < 0.0 {hfactor = 0.0}
				Ktotw := K + rho*Kw*hfactor

				// finally the fluxes -- the vertical
				// components include only the
				// physical K plus a weak numerical Kw
				// except for rain, which
				// needs additional damping in regions
				// where K == 0
				l.Fvxx[i] = -K*dxx
				l.Fvyy[i] = -K*dyy
				l.Fvzz[i] = -Ktotw*dzz

				l.Fvxy[i] = -K*dxy
				l.Fvyz[i] = -Ktotw*dyz
				l.Fvzx[i] = -Ktotw*dzx

				l.Fentx[i] = -K*dentx
				l.Fenty[i] = -K*denty
				l.Fentz[i] = -Ktotw*dentz + l.entdrrflux[iz]

				l.Frtx[i] = -K*drtx
				l.Frty[i] = -K*drty
				l.Frtz[i] = -Ktotw*drtz

				l.Frrx[i] = -K*drrx
				l.Frry[i] = -K*drry

				// include vertical numerical smoothing
				// and the fall of precip
				temp = thermo.Temp(l.irho[i],l.ip[i],
					l.idrt[i]/l.irho[i])
				wt := thermo.Ftrans(wtrain,wtsnow,temp)
				l.Frrz[i] = -Ktotz*drrz - wt*l.idrr[i]
			}

			// handle upper and lower boundaries
			ibb := iJ(ix,iy,0)
			itt := iJ(ix,iy,mz)
			ibb2 := iJ2(ix,iy)

			// set the eddy mixing coefficient to zero on boundaries
			l.ikeddy[ibb] = 0.0
			l.ikeddy[itt] = 0.0

			// set the boundary eddy fluxes to zero where
			// they are not controlled by surface flux
			// calculations
			l.Fvxx[ibb] = 0.0
			l.Fvxx[itt] = 0.0
			l.Fvyy[ibb] = 0.0
			l.Fvyy[itt] = 0.0
			l.Fvzz[ibb] = 0.0
			l.Fvzz[itt] = 0.0

			l.Fvxy[ibb] = 0.0
			l.Fvxy[itt] = 0.0
			l.Fvyz[ibb] = l.isurfvy[ibb2]
			l.Fvyz[itt] = 0.0
			l.Fvzx[ibb] = l.isurfvx[ibb2]
			l.Fvzx[itt] = 0.0

			l.Fentx[ibb] = 0.0
			l.Fentx[itt] = 0.0
			l.Fenty[ibb] = 0.0
			l.Fenty[itt] = 0.0
			l.Fentz[ibb] = l.isurfent[ibb2] - l.entdrrflux[0]
			l.Fentz[itt] = 0.0

			l.Frtx[ibb] = 0.0
			l.Frtx[itt] = 0.0
			l.Frty[ibb] = 0.0
			l.Frty[itt] = 0.0
			l.Frtz[ibb] = l.isurfrt[ibb2]
			l.Frtz[itt] = 0.0

			l.Frrx[ibb] = 0.0
			l.Frrx[itt] = 0.0
			l.Frry[ibb] = 0.0
			l.Frry[itt] = 0.0

			// include the sedimentation of precip
			temp := thermo.Temp(l.irho[ibb],l.ip[ibb],
				l.idrt[ibb]/l.irho[ibb])
			wt := thermo.Ftrans(wtrain,wtsnow,temp)
			l.Frrz[ibb] = - wt*l.idrr[ibb]
			l.Frrz[itt] = 0.0

			// extract the rain rate
			l.irain[ibb2] = - l.Frrz[ibb]
		}
	}

	// compute source terms on unstaggered grid
	for ix := 1; ix < mx + 1; ix++ {
		for iy := 1; iy < my + 1; iy++ {

			// compute radiative heating rate
			if radskp == 0 {  // fixed radiative heating profile
				srad := radiation(mz,dz,tp,radbrk,radheat)

				for iz := 0; iz < mz; iz++ {
					i := J(ix,iy,iz)
					
					// compute the temperature
					temp := thermo.Temp(l.rho[i],l.p[i],
						l.drt[i]/l.rho[i])
					
					// convert (K/s) to (W/K/m^3)
					l.sdrad[i] = l.rho[i]*thermo.Cp*srad[iz]/temp

					// zero the solar and net fluxes
					l.sflux[i] = 0.0
					l.flux[i] = 0.0
				}
			} else {  // interactive radiation

				// extract column data
				rt := make([]float64,mz)
				p := make([]float64,mz)
				rho := make([]float64,mz)
				rsat := make([]float64,mz)
				sent := make([]float64,mz)
				sflux := make([]float64,mz)
				flux := make([]float64,mz)
				for iz := 0; iz < mz; iz++ {
					i := J(ix,iy,iz)
					rt[iz] = l.rt[i]
					p[iz] = l.p[i]
					rho[iz] = l.rho[i]
					temp := thermo.Temp(rho[iz],p[iz],rt[iz])
					rsat[iz] = thermo.Rsat(temp,rho[iz])
				}
				stemp := l.sst[J2(ix,iy)]

				// call the toy radiation code
				toyrad.Radiation(mz,sun,cld,time,solarabs,solarwv,
					stemp,dz,rt,p,rho,rsat,sent,sflux,flux)

				// transfer entropy source
				for iz := 0; iz < mz; iz++ {
					i := J(ix,iy,iz)
					l.sdrad[i] = sent[iz]
					l.sflux[i] = sflux[iz]
					l.flux[i] = flux[iz]
				}
			}
				
			// interpolate eddy mixing coefficient
			// and eddy flux divergences to
			// unstaggered grid
			for iz := 0; iz < mz; iz++ {
				i := J(ix,iy,iz)
				l.keddy[i] = Ftr(ix,iy,iz,l.ikeddy[:])
				l.sux[i] = -(Fxd(ix,iy,iz,dx,l.Fvxx[:]) +
					Fyd(ix,iy,iz,dy,l.Fvxy[:]) +
					Fzd(ix,iy,iz,dz,l.Fvzx[:]))
				l.suy[i] = -(Fxd(ix,iy,iz,dx,l.Fvxy[:]) +
					Fyd(ix,iy,iz,dy,l.Fvyy[:]) +
					Fzd(ix,iy,iz,dz,l.Fvyz[:]))
				l.suz[i] = -(Fxd(ix,iy,iz,dx,l.Fvzx[:]) +
					Fyd(ix,iy,iz,dy,l.Fvyz[:]) +
					Fzd(ix,iy,iz,dz,l.Fvzz[:]))
				l.sdent[i] = -(Fxd(ix,iy,iz,dx,l.Fentx[:]) +
					Fyd(ix,iy,iz,dy,l.Fenty[:]) +
					Fzd(ix,iy,iz,dz,l.Fentz[:])) +
					l.sdrad[i]
				l.sdrt[i] = -(Fxd(ix,iy,iz,dx,l.Frtx[:]) +
					Fyd(ix,iy,iz,dy,l.Frty[:]) +
					Fzd(ix,iy,iz,dz,l.Frtz[:]))
				l.sdrr[i] = -(Fxd(ix,iy,iz,dx,l.Frrx[:]) +
					Fyd(ix,iy,iz,dy,l.Frry[:]) +
					Fzd(ix,iy,iz,dz,l.Frrz[:]))

				// compute the temperature
				temp := thermo.Temp(l.rho[i],l.p[i],
					l.drt[i]/l.rho[i])
				
				// compute temperature dependent
				// precip conversion term
				cprecip := thermo.Ftrans(crain,csnow,temp)

				// compute the rainfall generation rate
				newrain := 0.0
				rsat := thermo.Rsat(temp,l.rho[i])
				rc := l.rt[i] - rsat
				if rc > 0.0 {

					// saturated, so rain created
					newrain = cprecip*rc*dtcoarse

					// rate limit due to exhaustion of
					// condensate
					if newrain > rc {
						newrain = rc
					}
				} else {

					// unsaturated, so rain evaporated
					newrain = cevap*rc*l.rr[i]*dtcoarse

					// rate limit due to exhaustion of
					// rain
					if newrain < -l.rr[i] {
						newrain = -l.rr[i]
					}
				}

				// increment rain-related moisture source terms
				l.sdrt[i] -= newrain*l.rho[i]/dtcoarse
				l.sdrr[i] += newrain*l.rho[i]/dtcoarse

				// wtg contributions to rt and s
				l.sdrt[i] += wtg.srcrt[iz]
				l.sdent[i] += wtg.srcent[iz]

				// nudging contributions to ux and uy
				l.sux[i] += wtg.srcvx[iz]
				l.suy[i] += wtg.srcvy[iz]
			}

			// interpolate staggered surface fields to
			// unstaggered grid
			j := J2(ix,iy)
			l.surfvx[j] = Ftr2(ix,iy,l.isurfvx[:])
			l.surfvy[j] = Ftr2(ix,iy,l.isurfvy[:])
			l.surfent[j] = Ftr2(ix,iy,l.isurfent[:])
			l.surfrt[j] = Ftr2(ix,iy,l.isurfrt[:])
			l.rain[j] = Ftr2(ix,iy,l.irain[:])
			l.sst[j] = Ftr2(ix,iy,l.isst[:])
			l.surft[j] = Ftr2(ix,iy,l.isurft[:])
			l.surfentcum[j] += l.surfent[j]*dtcoarse
			l.surfrtcum[j] += l.surfrt[j]*dtcoarse
			l.raincum[j] += l.rain[j]*dtcoarse
		}
	}

	// accumulate WTG source term profiles
	for iz := 0; iz < mz; iz++ {
		wtg.srcrtcum[iz] += wtg.srcrt[iz]*dtcoarse
		wtg.srcentcum[iz] += wtg.srcent[iz]*dtcoarse
	}

	// do horizontal 2*dx-dy smoothing before incrementing
	// variables so that edge values are at same time level
	// as interior variables
	Chsmooth(lambdah,l.ux)
	Chsmooth(lambdah,l.uy)
	Chsmooth(lambdah,l.uz)
	Chsmooth(lambdah,l.dent)
	Chsmooth(lambdah,l.drt)
	Chsmooth(lambdah,l.drr)

	// increment variables
	for ix := 1; ix < mx + 1; ix++ {
		for iy := 1; iy < my + 1; iy++ {
			rentinit := ranent*(2.0*rand.Float64() - 1.0)
			for iz := 0; iz < mz; iz++ {
				i := J(ix,iy,iz)
				l.ux[i] += l.sux[i]*dtcoarse
				l.uy[i] += l.suy[i]*dtcoarse
				l.uz[i] += l.suz[i]*dtcoarse
				l.dent[i] += l.sdent[i]*dtcoarse
				l.drt[i] += l.sdrt[i]*dtcoarse
				l.drr[i] += l.sdrr[i]*dtcoarse

				// random perturbation of entropy field
				// done to excite convection in bistable
				// situations
				zval := (float64(iz) + 0.5)*dz
				l.dent[i] += rentinit*math.Exp(-zval/zscale)
			}
		}
	}

	// finish some details
	for ix := 1; ix < mx + 1; ix++ {
		for iy := 1; iy < my + 1; iy++ {
			for iz := 0; iz < mz; iz++ {
				i := J(ix,iy,iz)

				// don't let rain become negative --
				// we shouldn't have to worry about rt
				// becoming negative here, since rain
				// is mainly at low levels where rt
				// is large
				if l.drr[i] < 0.0 {
					l.drt[i] += l.drr[i]
					l.drr[i] = 0.0
				}

				// diagnose pressure
				l.p[i] = thermo.Pres2(l.rho[i],
					l.drt[i]/l.rho[i],l.drr[i]/l.rho[i],
					l.dent[i]/l.rho[i])
			}
		}
	}

	// do edge transfers to propagate modified interior values to
	// edges
	edgetrans(comm,rank,gx,gy,l)
}

// *******************************************************************
// compute fixed radiative heating rate (K/s) as a function of height
func radiation(mz int, dz, tp, radbrk, radheat float64) (srad []float64) {

	// make a temporary radiation heating profile
	srad = make([]float64,mz)

	// compute fixed radiative heating
	for iz := 0; iz < mz; iz++ {
		z := float64(iz)*dz
		prof := (tp - z)/(tp - radbrk)
		if prof > 1.0 {prof = 1.0}
		if prof < 0.0 {prof = 0.0}
		srad[iz] = radheat*prof
	}
	return
}

// *******************************************************************
// *******************************************************************
// *******************************************************************
// *******************************************************************
// *******************************************************************
// do weak temperature gradient and related calculations
func wtgcalc(nx, ny, nz int, g *gfields, wtg *wtgfields,
	paramlist map[string]string) {
	
	// get wtg-related parameters
	dz := params.Fparam(paramlist,"dz")
	drelax := params.Fparam(paramlist,"drelax")
	trelax := params.Fparam(paramlist,"trelax")
	mingrad := params.Fparam(paramlist,"mingrad")
	f0 := params.Fparam(paramlist,"f0")
	spectral := params.Iparam(paramlist,"spectral")
	kappa := params.Fparam(paramlist,"kappa")
	sobel := params.Iparam(paramlist,"sobel")
	srelax := params.Fparam(paramlist,"srelax")
	bltop := params.Fparam(paramlist,"bltop")

	// define local profiles
	diff := make([]float64,nz)
	grad := make([]float64,nz)
	dryent := make([]float64,nz)
	wtgdryent := make([]float64,nz)
	ent := make([]float64,nz)
	rt := make([]float64,nz)
	vz := make([]float64,nz)
	mconv := make([]float64,nz)
	wtguzstar := make([]float64,nz)
	
	// increment source terms for wtg
	for iz := 0; iz < nz; iz++ {

		// concentrations from densities
		dryent[iz] = wtg.ddryentmean[iz]/wtg.rhomean[iz]
		ent[iz] = wtg.dentmean[iz]/wtg.rhomean[iz]
		rt[iz] = wtg.drtmean[iz]/wtg.rhomean[iz]
		wtgdryent[iz] = wtg.ddryent[iz]/wtg.rho[iz]
		
		// compute dry entropy wtg deviation
		diff[iz] = dryent[iz] - wtgdryent[iz]
	}
	
	// compute vertical gradient of dryent
	for iz := 1; iz < nz - 1; iz++ {
		grad[iz] = (dryent[iz + 1] - dryent[iz - 1])/(2.0*dz)
		if grad[iz] < mingrad {
			grad[iz] = mingrad
		}
	}
	
	// off-centered differencing at end points
	grad[0] = (dryent[1] - dryent[0])/dz
	grad[nz - 1] = (dryent[nz - 1] - dryent[nz - 2])/dz
	
	// compute the Fourier coefficients
	wtg.dryentj[0] = 0.0
	for j := 1; j < nj; j++ {
		wtg.dryentj[j] = 0.0

		// select spectral or classic wtg
		trelaxj := trelax
		if spectral == 1 {
			trelaxj = trelaxj/float64(j)
		}

		// loop over column for given value of j
		for iz := 0; iz < nj; iz++ {

			// if trelaxj and f0 both equal zero, this is
			// a special case in which omega = 0
			omega := trelaxj*trelaxj
			denom := f0*f0 + trelaxj*trelaxj
			if denom > 0.0 {
				omega = omega/math.Pow(denom,0.5)
			}
			delta := diff[iz]/grad[iz]
			wtg.dryentj[j] += omega*delta*wtg.sine0[SI(iz,j)]
		}
		wtg.dryentj[j] *= 2.0/float64(nj)
	}
	
	// compute wtg vertical mass flux
	for iz := nz - 1; iz >= 0; iz-- {
		if iz >= nj {

			// zeroed above tropopause
			wtg.uz[iz] = 0.0
		} else {
			
			// sum wtg.dryentj-weighted sine functions at each
			// level -- j == 0 is trivial
			vz[iz] = 0.0
			for j := 1; j < nj; j++ {
				vz[iz] += wtg.dryentj[j]*wtg.sine0[SI(iz,j)]
			}
			wtg.uz[iz] = vz[iz]*wtg.rho[iz]
		}
	}

	// rationalize wtg.uz in boundary layer
	ibl := int(bltop/dz - 0.5)
	if ibl < 0 {ibl = 0}
	if ibl > nz - 1 {ibl = nz - 1}
	for iz := 0; iz < ibl; iz++ {
		wtg.uz[iz] = wtg.uz[ibl]*g.z[iz]/g.z[ibl]
	}

	// do 1-2-1 filtering on wtg.uz
	for iz := 1; iz < nz - 1; iz++ {
		wtguzstar[iz] = 0.5*wtg.uz[iz] + 0.25*(wtg.uz[iz + 1] +
			wtg.uz[iz - 1])
	}
	for iz := 1; iz < nz - 1; iz++ {
		wtg.uz[iz] = wtguzstar[iz]
	}
	
	// two different treatments of entropy/moisture entrainment
	// and detrainment
	if sobel == 1 {  // Wang-Sobel treatment
		for iz := 0; iz < nz; iz++ {
			wtg.srcent[iz] = srelax*(wtg.dent[iz] -
				wtg.dentmean[iz])
			wtg.srcrt[iz] = srelax*(wtg.drt[iz] -
				wtg.drtmean[iz])
		}
	} else { // Raymond-Zeng treatment
	
		// compute divergence term using centered differencing
		// off-centered differencing at top and bottom
		for iz := 1; iz < nz - 1; iz++ {
			mconv[iz] = (wtg.uz[iz + 1] - wtg.uz[iz - 1])/
				(2.0*wtg.rho[iz]*dz)
		}
		mconv[0] = (wtg.uz[1] - wtg.uz[0])/(wtg.rho[0]*dz)
		mconv[nz - 1] = (wtg.uz[nz - 1] - wtg.uz[nz - 2])/
			(wtg.rho[nz - 1]*dz)

		// zero source terms and add effect of ventilation to mconv
		for iz := 0; iz < nz; iz++ {
			wtg.srcrt[iz] = 0.0
			wtg.srcent[iz] = 0.0
			uhsq := wtg.ux[iz]*wtg.ux[iz] + wtg.uy[iz]*wtg.uy[iz]
			mconv[iz] += kappa*math.Pow(uhsq,0.5)
		}
		
		// vertical advection of moisture and entropy -- centered diff
		for iz := 1; iz < nz - 1; iz++ {
			wtg.srcrt[iz] -= wtg.uz[iz]*
				(rt[iz + 1] - rt[iz - 1])/(2.0*dz)
			wtg.srcent[iz] -= wtg.uz[iz]*
				(ent[iz + 1] - ent[iz - 1])/(2.0*dz)
		}
	
		// lateral entrainment term
		for iz := 0; iz < nz; iz++ {
			if mconv[iz] > 0.0 {
				wtg.srcrt[iz] -= (wtg.drtmean[iz] -
					wtg.drt[iz])*mconv[iz]
				wtg.srcent[iz] -= (wtg.dentmean[iz] -
					wtg.dent[iz])*mconv[iz]
			}
		}
	}
	
	// compute velocity nudging source terms
	for iz := 0; iz < nz; iz++ {
		r := wtg.rho[iz]/wtg.rhomean[iz]
		wtg.srcvx[iz] = drelax*(wtg.ux[iz] - wtg.uxmean[iz]*r)
		wtg.srcvy[iz] = drelax*(wtg.uy[iz] - wtg.uymean[iz]*r)
	}
}

// *******************************************************************
func gatherwtg(comm gompi.MPI_Comm, nz, rank, gx, gy int, l *lfields,
	wtg *wtgfields) {
	
	// compute horizontal averages and add to send buffer
	count := 0
	count = wtgpack(count,l.ux[:],wtg.sgbuff[:])
	count = wtgpack(count,l.uy[:],wtg.sgbuff[:])
	count = wtgpack(count,l.rho[:],wtg.sgbuff[:])
	count = wtgpack(count,l.drt[:],wtg.sgbuff[:])
	count = wtgpack(count,l.ddryent[:],wtg.sgbuff[:])
	count = wtgpack(count,l.dent[:],wtg.sgbuff[:])
	psize := meanfields*nz

	// simple copy if rank == 0, otherwise MPI send
	gwtgflag := make(chan int,0)
	if rank == 0 {
		for i := 0; i < psize; i++ {
			wtg.rgbuff[i] = wtg.sgbuff[i]
		}
	} else {
		go func() {
			gompi.Send_float64(wtg.sgbuff[:],0,comm)
			gwtgflag <- 0
		} ()
	}

	// at the receiving end...
	if rank == 0 {
	
		// zero the mean fields
		for iz := 0; iz < nz; iz++ {
			wtg.uxmean[iz] = 0.0
			wtg.uymean[iz] = 0.0
			wtg.rhomean[iz] = 0.0
			wtg.drtmean[iz] = 0.0
			wtg.ddryentmean[iz] = 0.0
			wtg.dentmean[iz] = 0.0
		}
	
		// receiver has to loop over all senders except rank == 0
		for xrank := 1; xrank < gx*gy; xrank++ {

			// get the results from rank == xrank
			p1 := xrank*psize
			p2 := (xrank + 1)*psize
			gompi.Recv_float64(wtg.rgbuff[p1:p2],xrank,comm)
		}

		// loop over senders again to extract data
		count = 0
		scale := 1.0/float64(gx*gy)
		for xrank := 0; xrank < gx*gy; xrank++ {

			// do the sums for averaging
			for iz := 0; iz < nz; iz++ {
				wtg.uxmean[iz] += wtg.rgbuff[count]*scale
				count += 1
			}
			for iz := 0; iz < nz; iz++ {
				wtg.uymean[iz] += wtg.rgbuff[count]*scale
				count += 1
			}
			for iz := 0; iz < nz; iz++ {
				wtg.rhomean[iz] += wtg.rgbuff[count]*scale
				count += 1
			}
			for iz := 0; iz < nz; iz++ {
				wtg.drtmean[iz] += wtg.rgbuff[count]*scale
				count += 1
			}
			for iz := 0; iz < nz; iz++ {
				wtg.ddryentmean[iz] += wtg.rgbuff[count]*scale
				count += 1
			}
			for iz := 0; iz < nz; iz++ {
				wtg.dentmean[iz] += wtg.rgbuff[count]*scale
				count += 1
			}
		}
	} else {

	// finished with sends from ranks > 0
		<- gwtgflag
	}
}

// *******************************************************************
// pack 3D fields for WTG gather
func wtgpack(count int, lfield, buff []float64) (newcount int) {

	// set up buffer index
	newcount = count

	// compute horizontal sum
	for iz := 0; iz < mz; iz++ {
		sum := 0.0
		for ix := 0; ix < mx; ix++ {
			for iy := 0; iy < my; iy++ {
				j := J(ix + 1,iy + 1,iz)
				sum = sum + lfield[j]
			}
		}
		sum = sum/float64(mx*my)

		// make transfer to buffer and increment pointer
		buff[newcount] = sum
		newcount += 1
	}

	// return updated pointer
	return
}

// *******************************************************************
// scatter wtg terms to all other ranks if needed
func scatterwtg(comm gompi.MPI_Comm, nz, rank, gx, gy int, wtg *wtgfields) {

	// send stuff from rank == 0 to other ranks
	if rank == 0 && gx*gy > 1 {

		// fill buffer
		count := 0
		for iz := 0; iz < nz; iz++ {
			wtg.ssbuff[count] = wtg.srcvx[iz]
			count += 1
			wtg.ssbuff[count] = wtg.srcvy[iz]
			count += 1
			wtg.ssbuff[count] = wtg.srcrt[iz]
			count += 1
			wtg.ssbuff[count] = wtg.srcent[iz]
			count += 1
		}

		// don't need a goroutine as ranks don't talk to self
		for xrank := 1; xrank < gx*gy; xrank++ {
			gompi.Send_float64(wtg.ssbuff[:],xrank,comm)
		}
	}

	// do the receive for other than rank == 0
	if rank > 0 {

		// blocking receive
		gompi.Recv_float64(wtg.rsbuff[:],0,comm)
		
		// unpack the source fields from the received buffer
		count := 0
		for iz := 0; iz < nz; iz++ {
			wtg.srcvx[iz] = wtg.rsbuff[count]
			count += 1
			wtg.srcvy[iz] = wtg.rsbuff[count]
			count += 1
			wtg.srcrt[iz] = wtg.rsbuff[count]
			count += 1
			wtg.srcent[iz] = wtg.rsbuff[count]
			count += 1
		}
	}
}

// *******************************************************************
// *******************************************************************
// *******************************************************************
// *******************************************************************
// write an output file
func docandis(pfile string, nx, ny, nz, seq int, g *gfields,
	wtg *wtgfields, paramlist map[string]string) {

	// get parameters
	mt := params.Iparam(paramlist,"mt")  // coarse time loops
	lt := params.Iparam(paramlist,"lt")  // fine time loops
	dt := params.Fparam(paramlist,"dt")  // fine scale timestep (s)

	// calculate coarse time
	time := float64(seq*mt*lt)*dt

	// convert seq to a zero-filled int and create a file name
	filename := fmt.Sprintf("%s.%04d",pfile,seq)

	// turn time, vxsys, and vysys into slices
	stime := make([]float64,1)
	stime[0] = time
	svxsys := make([]float64,1)
	svxsys[0] = wtg.vxsys
	svysys := make([]float64,1)
	svysys[0] = wtg.vysys

	// create 3D diagnostic fields
	temp := make([]float64,nsize)  // temperature
	drsat := make([]float64,nsize)  // saturation mr*rho
	ddryent := make([]float64,nsize)  // dryent*rho
	dsatent := make([]float64,nsize)  // satent*rho
	for i := 0; i < nsize; i++ {
		temp[i] = thermo.Temp(g.rho[i],g.p[i],g.drt[i]/g.rho[i])
		rsat := thermo.Rsat(temp[i],g.rho[i])
		drsat[i] = rsat*g.rho[i]
		dryent := thermo.Dent(temp[i],g.rho[i],g.drt[i]/g.rho[i])
		ddryent[i] = dryent*g.rho[i]
		satent := thermo.Sent(temp[i],g.rho[i])
		dsatent[i] = satent*g.rho[i]
	}

	// now some 1D diagnostic fields
	wtgrt := make([]float64,nz)
	wtgdryent := make([]float64,nz)
	wtgent := make([]float64,nz)
	wtgsatent := make([]float64,nz)
	wtgvx := make([]float64,nz)
	wtgvy := make([]float64,nz)
	for iz := 0; iz < nz; iz++ {
		rho := wtg.rho[iz]
		wtgrt[iz] = wtg.drt[iz]/rho
		wtgdryent[iz] = wtg.ddryent[iz]/rho
		wtgent[iz] = wtg.dent[iz]/rho
		wtgsatent[iz] = wtg.dsatent[iz]/rho
		wtgvx[iz] = wtg.ux[iz]/rho
		wtgvy[iz] = wtg.uy[iz]/rho
	}

	// create and write the candis file
	a := gocandis.NullCandis()
	a.AddComment("Goconv " + pfile)
	a.AddComment("Version-" + VERSION)
	a.AddComment(fmt.Sprintf("mx = %d, my = %d, mz = %d",mx,my,mz))
	a.AddComment(fmt.Sprintf("gx = %d, gy = %d",gx,gy))
	a.AddComment(fmt.Sprintf("mt = %d, lt = %d, dt = %g",mt,lt,dt))
	a.AddDim("x",g.x[:])
	a.AddDim("y",g.y[:])
	a.AddDim("z",g.z[:])
	a.AddVField("time",[]string{},stime)
	a.AddVField("vxsys",[]string{},svxsys)
	a.AddVField("vysys",[]string{},svysys)
	a.AddVField("wtgux",[]string{"z"},wtg.ux[:])
	a.AddVField("wtguy",[]string{"z"},wtg.uy[:])
	a.AddVField("wtgpres",[]string{"z"},wtg.pres[:])
	a.AddVField("wtgrho",[]string{"z"},wtg.rho[:])
	a.AddVField("wtgdrt",[]string{"z"},wtg.drt[:])
	a.AddVField("wtgddryent",[]string{"z"},wtg.ddryent[:])
	a.AddVField("wtgdent",[]string{"z"},wtg.dent[:])
	a.AddVField("wtgdsatent",[]string{"z"},wtg.dsatent[:])
	a.AddVField("wtguz",[]string{"z"},wtg.uz[:])
	a.AddVField("wtgrt",[]string{"z"},wtgrt[:])
	a.AddVField("wtgdryent",[]string{"z"},wtgdryent[:])
	a.AddVField("wtgent",[]string{"z"},wtgent[:])
	a.AddVField("wtgsatent",[]string{"z"},wtgsatent[:])
	a.AddVField("wtgvx",[]string{"z"},wtgvx[:])
	a.AddVField("wtgvy",[]string{"z"},wtgvy[:])
	a.AddVField("wtgsrcvx",[]string{"z"},wtg.srcvx[:])
	a.AddVField("wtgsrcvy",[]string{"z"},wtg.srcvy[:])
	a.AddVField("wtgsrcrt",[]string{"z"},wtg.srcrt[:])
	a.AddVField("wtgsrcent",[]string{"z"},wtg.srcent[:])
	a.AddVField("wtgsrcrtcum",[]string{"z"},wtg.srcrtcum[:])
	a.AddVField("wtgsrcentcum",[]string{"z"},wtg.srcentcum[:])
	a.AddVField("rho",[]string{"x","y","z"},g.rho[:])
	a.AddVField("pres",[]string{"x","y","z"},g.p[:])
	a.AddVField("ux",[]string{"x","y","z"},g.ux[:])
	a.AddVField("uy",[]string{"x","y","z"},g.uy[:])
	a.AddVField("uz",[]string{"x","y","z"},g.uz[:])
	a.AddVField("ddryent",[]string{"x","y","z"},ddryent)
	a.AddVField("dent",[]string{"x","y","z"},g.dent[:])
	a.AddVField("dsatent",[]string{"x","y","z"},dsatent)
	a.AddVField("drt",[]string{"x","y","z"},g.drt[:])
	a.AddVField("drr",[]string{"x","y","z"},g.drr[:])
	a.AddVField("drsat",[]string{"x","y","z"},drsat)
	a.AddVField("temp",[]string{"x","y","z"},temp)
	a.AddVField("keddy",[]string{"x","y","z"},g.keddy[:])
	a.AddVField("sdrad",[]string{"x","y","z"},g.sdrad[:])
	a.AddVField("sflux",[]string{"x","y","z"},g.sflux[:])
	a.AddVField("flux",[]string{"x","y","z"},g.flux[:])
	a.AddVField("sux",[]string{"x","y","z"},g.sux[:])
	a.AddVField("suy",[]string{"x","y","z"},g.suy[:])
	a.AddVField("sdent",[]string{"x","y","z"},g.sdent[:])
	a.AddVField("sdrt",[]string{"x","y","z"},g.sdrt[:])
	a.AddVField("sdrr",[]string{"x","y","z"},g.sdrr[:])
	a.AddVField("surfvx",[]string{"x","y"},g.surfvx[:])
	a.AddVField("surfvy",[]string{"x","y"},g.surfvy[:])
	a.AddVField("surfent",[]string{"x","y"},g.surfent[:])
	a.AddVField("surfrt",[]string{"x","y"},g.surfrt[:])
	a.AddVField("rain",[]string{"x","y"},g.rain[:])
	a.AddVField("surfentcum",[]string{"x","y"},g.surfentcum[:])
	a.AddVField("surfrtcum",[]string{"x","y"},g.surfrtcum[:])
	a.AddVField("raincum",[]string{"x","y"},g.raincum[:])
	a.AddVField("sst",[]string{"x","y"},g.sst[:])
	a.AddVField("surft",[]string{"x","y"},g.surft[:])
	a.PutCandis(filename)
	a.CloseCandis()
}

// *******************************************************************
// divide field by density field
func dbrho(dfield, rhofld []float64) (field []float64) {
	length := len(dfield)
	length2 := len(rhofld)
	if length != length2 {
		panic("field length discrepancy in dbrho")
	}
	field = make([]float64, length)
	for i := 0; i < length; i++ {
		field[i] = dfield[i]/rhofld[i]
	}
	return
}

// *******************************************************************
// *******************************************************************
// *******************************************************************
// *******************************************************************
// *******************************************************************
// index calculators for both local and global 3D fields
// iz iterates most rapidly followed by iy followed by ix
// thus vertical columns have adjacent elements
func I(ix, iy, iz int) (i int) {
	i = iz + nz*(iy + ny*ix)
	return
}

// 2D global field
func I2(ix, iy int) (i int) {
	i = iy + ny*ix
	return
}

// unstaggered grid
func J(ix, iy, iz int) (i int) {
	i = iz + lz*(iy + ly*ix)
	return
}

// 2D unstaggered grid
func J2(ix, iy int) (i int) {
	i = iy + ly*ix
	return
}

// staggered grid
func iJ(ix, iy, iz int) (i int) {
	i = iz + ilz*(iy + ily*ix)
	return
}

// 2D staggered grid
func iJ2(ix, iy int) (i int) {
	i = iy + ily*ix
	return
}

// for WTG
func SI(iz, j int) (wtgindex int) {
	wtgindex = j + nj*iz
	return
}

// *******************************************************************
// locate lower left corner of local array in global array given rank
// values of xloc run from 0 to gx - 1
// values of yloc run from 0 to gy - 1
// the y dimension iterates most rapidly with increasing rank
func L(rank, gx, gy int) (xloc, yloc int) {
	xloc = rank/gy
	yloc = rank - gy*xloc
	return
}

// *******************************************************************
// find the rank of a local array with specified xloc and yloc
// rolls over for +-1 out of bound xloc and yloc
func M(xloc, yloc, gx, gy int) (rank int) {
	if xloc < 0 {xloc += gx}
	if xloc >= gx {xloc -= gx}
	if yloc < 0 {yloc += gy}
	if yloc >= gy {yloc -= gy}
	rank = yloc + gy*xloc
	return
}

// *******************************************************************
// compute staggered value from unstaggered values
func Ctr(ix, iy, iz int, lfield []float64) (ilvalue float64) {
	ilvalue = 0.125*(
		lfield[J(ix + 1,iy + 1,iz)] +
		lfield[J(ix,iy + 1,iz)] +
		lfield[J(ix + 1,iy,iz)] +
		lfield[J(ix,iy,iz)] +
		lfield[J(ix + 1,iy + 1,iz - 1)] +
		lfield[J(ix,iy + 1,iz - 1)] +
		lfield[J(ix + 1,iy,iz - 1)] +
		lfield[J(ix,iy,iz - 1)])
	return
}

// *******************************************************************
// compute 2D staggered value from unstaggered values
func Ctr2(ix, iy int, lfield []float64) (ilvalue float64) {
	ilvalue = 0.25*(
		lfield[J2(ix + 1,iy + 1)] +
		lfield[J2(ix,iy + 1)] +
		lfield[J2(ix + 1,iy)] +
		lfield[J2(ix,iy)])
	return
}

// *******************************************************************
// horizontal smoothing operator for 3D unstaggered grid -- doesn't apply to
// boundaries -- range on lambdah: 0 <= lambdah <= 1
func Chsmooth(lambdah float64, lf []float64) {

	// make temporary smoothed field
	lfs := make([]float64, lsize)

	// create filtered interior points, not changing boundary values
	for iz := 0; iz < mz; iz++ {
		for ix := 1; ix < lx - 1; ix++ {
			for iy := 1; iy < ly - 1; iy++ {
				j := J(ix,iy,iz)
				jpp := J(ix + 1,iy + 1,iz)
				jpm := J(ix + 1,iy - 1,iz)
				jmp := J(ix - 1,iy + 1,iz)
				jmm := J(ix - 1,iy - 1,iz)
				lfs[j] = (1.0 - lambdah)*lf[j] +
					0.25*lambdah*(lf[jpp] + lf[jpm] +
					lf[jmp] + lf[jmm])
			}
		}
	}

	// transfer smoothed interior values back to original field
	for iz := 0; iz < mz; iz++ {
		for ix := 1; ix < lx - 1; ix++ {
			for iy := 1; iy < ly - 1; iy++ {
				j := J(ix,iy,iz)
				lf[j] = lfs[j]
			}
		}
	}
}

// *******************************************************************
// compute staggered x derivative of lf from unstaggered values
func Cxd(ix, iy, iz int, dx float64, lf []float64) (ilv float64) {
	ippp := J(ix + 1,iy + 1,iz)
	inpp := J(ix,iy + 1,iz)
	ipnp := J(ix + 1,iy,iz)
	ippn := J(ix + 1,iy + 1,iz - 1)
	innp := J(ix,iy,iz)
	ipnn := J(ix + 1,iy,iz - 1)
	inpn := J(ix,iy + 1,iz - 1)
	innn := J(ix,iy,iz - 1)
	ilv = 0.25*(
		lf[ippp] +
		lf[ipnp] +
		lf[ippn] +
		lf[ipnn] -
		lf[inpp] -
		lf[innp] -
		lf[inpn] -
		lf[innn])/dx
	return
}

// *******************************************************************
// compute staggered y derivative of lf from unstaggered values
func Cyd(ix, iy, iz int, dy float64, lf []float64) (ilv float64) {
	ippp := J(ix + 1,iy + 1,iz)
	inpp := J(ix,iy + 1,iz)
	ipnp := J(ix + 1,iy,iz)
	ippn := J(ix + 1,iy + 1,iz - 1)
	innp := J(ix,iy,iz)
	ipnn := J(ix + 1,iy,iz - 1)
	inpn := J(ix,iy + 1,iz - 1)
	innn := J(ix,iy,iz - 1)
	ilv = 0.25*(
		lf[ippp] -
		lf[ipnp] +
		lf[ippn] -
		lf[ipnn] +
		lf[inpp] -
		lf[innp] +
		lf[inpn] -
		lf[innn])/dy
	return
}

// *******************************************************************
// compute staggered z derivative of lf from unstaggered values
func Czd(ix, iy, iz int, dz float64, lf []float64) (ilv float64) {
	ippp := J(ix + 1,iy + 1,iz)
	inpp := J(ix,iy + 1,iz)
	ipnp := J(ix + 1,iy,iz)
	ippn := J(ix + 1,iy + 1,iz - 1)
	innp := J(ix,iy,iz)
	ipnn := J(ix + 1,iy,iz - 1)
	inpn := J(ix,iy + 1,iz - 1)
	innn := J(ix,iy,iz - 1)
	ilv = 0.25*(
		lf[ippp] +
		lf[ipnp] -
		lf[ippn] -
		lf[ipnn] +
		lf[inpp] +
		lf[innp] -
		lf[inpn] -
		lf[innn])/dz
	return
}

// *******************************************************************
// compute unstaggered value from staggered values
func Ftr(ix, iy, iz int, ilfield []float64) (lvalue float64) {
	lvalue = 0.125*(
		ilfield[iJ(ix,iy,iz)] +
		ilfield[iJ(ix - 1,iy,iz)] +
		ilfield[iJ(ix,iy - 1,iz)] +
		ilfield[iJ(ix - 1,iy - 1,iz)] +
		ilfield[iJ(ix,iy,iz + 1)] +
		ilfield[iJ(ix - 1,iy,iz + 1)] +
		ilfield[iJ(ix,iy - 1,iz + 1)] +
		ilfield[iJ(ix - 1,iy - 1,iz + 1)])
	return
}

// *******************************************************************
// compute 2D unstaggered value from staggered values
func Ftr2(ix, iy int, ilfield []float64) (lvalue float64) {
	lvalue = 0.25*(
		ilfield[iJ2(ix,iy)] +
		ilfield[iJ2(ix - 1,iy)] +
		ilfield[iJ2(ix,iy - 1)] +
		ilfield[iJ2(ix - 1,iy - 1)])
	return
}

// *******************************************************************
// compute final x derivative of lif from staggered values
func Fxd(ix, iy, iz int, dx float64, lif []float64) (lv float64) {
	ippp := iJ(ix,iy,iz + 1)
	inpp := iJ(ix - 1,iy,iz + 1)
	ipnp := iJ(ix,iy - 1,iz + 1)
	innp := iJ(ix - 1,iy - 1,iz + 1)
	ippn := iJ(ix,iy,iz)
	ipnn := iJ(ix,iy - 1,iz)
	inpn := iJ(ix - 1,iy,iz)
	innn := iJ(ix - 1,iy - 1,iz)
	lv = 0.25*(
		lif[ippp] +
		lif[ipnp] +
		lif[ippn] +
		lif[ipnn] -
		lif[inpp] -
		lif[innp] -
		lif[inpn] -
		lif[innn])/dx
	return
}

// *******************************************************************
// compute final y derivative of lif from staggered values
func Fyd(ix, iy, iz int, dy float64, lif []float64) (lv float64) {
	ippp := iJ(ix,iy,iz + 1)
	inpp := iJ(ix - 1,iy,iz + 1)
	ipnp := iJ(ix,iy - 1,iz + 1)
	innp := iJ(ix - 1,iy - 1,iz + 1)
	ippn := iJ(ix,iy,iz)
	ipnn := iJ(ix,iy - 1,iz)
	inpn := iJ(ix - 1,iy,iz)
	innn := iJ(ix - 1,iy - 1,iz)
	lv = 0.25*(
		lif[ippp] -
		lif[ipnp] +
		lif[ippn] -
		lif[ipnn] +
		lif[inpp] -
		lif[innp] +
		lif[inpn] -
		lif[innn])/dy
	return
}

// *******************************************************************
// compute final z derivative of lif from staggered values
func Fzd(ix, iy, iz int, dz float64, lif []float64) (lv float64) {
	ippp := iJ(ix,iy,iz + 1)
	inpp := iJ(ix - 1,iy,iz + 1)
	ipnp := iJ(ix,iy - 1,iz + 1)
	innp := iJ(ix - 1,iy - 1,iz + 1)
	ippn := iJ(ix,iy,iz)
	ipnn := iJ(ix,iy - 1,iz)
	inpn := iJ(ix - 1,iy,iz)
	innn := iJ(ix - 1,iy - 1,iz)
	lv = 0.25*(
		lif[ippp] +
		lif[ipnp] -
		lif[ippn] -
		lif[ipnn] +
		lif[inpp] +
		lif[innp] -
		lif[inpn] -
		lif[innn])/dz
	return
}

// local dynamical core fields
type lfields struct {

	// constant SST value
	sst0 float64

	// local unstaggered grid fields
	rho []float64    // density (kg/m^3)
	ux []float64     // x velocity*rho (kg/m^2/s)
	uy []float64     // y velocity*rho
	uz []float64     // z velocity*rho
	dent []float64   // specific entropy*rho (J/K/m^3)
	p []float64      // pressure (Pa)
	drt []float64    // total cw mixing ratio*rho (kg/m^3)
	drr []float64    // precip mixing ratio*rho (kg/m^3)
	keddy []float64  // eddy mixing coefficient (kg/m/s)

	// accumulated negative and in time step
	drtnegative float64
	drttotal float64

	// local staggered grid fields
	irho []float64
	iux []float64
	iuy []float64
	iuz []float64
	ident []float64
	ip []float64
	idrt []float64
	idrr []float64
	ikeddy []float64

	// staggered 1D column of precip entropy flux
	entdrrflux []float64

	// source terms on unstaggered grid
	sux []float64
	suy []float64
	suz []float64
	sdent []float64
	sdrt []float64
	sdrr []float64
	sdrad []float64
	sflux []float64
	flux []float64
	
	// unstaggered fields divided by density
	vx []float64
	vy []float64
	vz []float64
	ent []float64
	rt []float64
	rr []float64

	// flux buffers (both staggered and
	// unstaggered grids)
	fxx []float64
	fyy []float64
	fzz []float64
	fxy []float64
	fyz []float64
	fzx []float64
	fxent []float64
	fyent []float64
	fzent []float64
	fxrt []float64
	fyrt []float64
	fzrt []float64
	fxrr []float64
	fyrr []float64
	fzrr []float64

	// coarse timestep flux buffers
	Fvxx []float64
	Fvyy []float64
	Fvzz []float64
	Fvxy []float64
	Fvyz []float64
	Fvzx []float64
	Fentx []float64
	Fenty []float64
	Fentz []float64
	Frtx []float64
	Frty []float64
	Frtz []float64
	Frrx []float64
	Frry []float64
	Frrz []float64

	// also need dry entropy on unstaggered grid
	// this is computed in coarse() and is needed
	// in wtgcalc()
	dryent []float64
	ddryent []float64

	// surface fields on unstaggered grid
	surfvx []float64
	surfvy []float64
	surfent []float64
	surfrt []float64
	rain []float64
	surfentcum []float64
	surfrtcum []float64
	raincum []float64
	sst []float64
	surft []float64

	// surface fields on staggered grid
	isurfvx []float64
	isurfvy []float64
	isurfent []float64
	isurfrt []float64
	irain []float64
	isst []float64
	isurft []float64

	// buffers for MPI transfers so that
	// processors can talk to nearest neighbors
	sbuffn []float64
	rbuffn []float64
	sbuffs []float64
	rbuffs []float64
	sbuffe []float64
	rbuffe []float64
	sbuffw []float64
	rbuffw []float64
	sbuffne []float64
	rbuffne []float64
	sbuffnw []float64
	rbuffnw []float64
	sbuffsw []float64
	rbuffsw []float64
	sbuffse []float64
	rbuffse []float64

	// scatter/gather buffers
	buff []float64
	rbuff []float64
}	

// global fields
type gfields struct {
	x []float64
	y []float64
	z []float64
	rho []float64
	ux []float64
	uy []float64
	uz []float64
	dent []float64
	p []float64
	drt []float64
	drr []float64
	sux []float64
	suy []float64
	suz []float64
	sdent []float64
	sdrt []float64
	sdrr []float64
	keddy []float64
	sdrad []float64
	sflux []float64
	flux []float64
	surfvx []float64
	surfvy []float64
	surfent []float64
	surfrt []float64
	rain []float64
	surfentcum []float64
	surfrtcum []float64
	raincum []float64
	sst []float64
	surft []float64
}

// wtg fields -- these are single column reference
// profiles and source terms
type wtgfields struct {

	// wtg reference profiles
	ux []float64
	uy []float64
	pres []float64
	rho []float64
	drt []float64
	ddryent []float64
	dent []float64
	dsatent []float64

	// domain velocity relative to earth
	vxsys float64
	vysys float64

	// wtg vertical mass flux
	uz []float64

	// wtg mean fields for wtg calcs
	uxmean []float64
	uymean []float64
	rhomean []float64
	drtmean []float64
	ddryentmean []float64
	dentmean []float64

	// wtg source terms needed in general calculation
	srcvx []float64
	srcvy []float64
	srcrt []float64
	srcent []float64
	srcrtcum []float64
	srcentcum []float64

	// MPI buffers
	sgbuff []float64
	rgbuff []float64
	ssbuff []float64
	rsbuff []float64

	// wtg calculation
	dryentj []float64
	sine0 []float64
	
	// stratospheric damping profile
	mu []float64

}
