// gosounding -- extrapolate an observed sounding upward with standard
// procedure.
// Input file assumes z in (m) with uniform spacing dz, starting at
// dz/2.  The input fields are pressure (hPa), temperature (C),
// mixing ratio (g/kg), and earth-relative wind (m/s).
package main

// packages needed
import (
	"gocandis"
	"thermo"
	"os"
	"fmt"
	"strconv"
)

const (
	
	// accel of gravity (m/s^2)
	Gee = 9.8

	// height of tropopause (m)
	Ztrop = 15000.0
)

func main() {

	// get command line arguments
	args := os.Args
	if len(args) != 5 {
		fmt.Println("Usage: gosounding sst vxsys vysys obtop")
		os.Exit(1)
	}

	// sea surface temp (K)
	sstin, _ := strconv.ParseFloat(args[1],64)
	sst := make([]float64, 1)
	sst[0] = sstin

	// x and y components of assumed domain propagation velocity
	vxsysin, _ := strconv.ParseFloat(args[2],64)
	vxsys := make([]float64, 1)
	vxsys[0] = vxsysin
	vysysin, _ := strconv.ParseFloat(args[3],64)
	vysys := make([]float64, 1)
	vysys[0] = vysysin
	Obtop, _ := strconv.ParseFloat(args[4],64)

	// initialize thermodynamics with ice turned on
	thermo.Init(1)

	// get the original sounding from standard input
	a := gocandis.GetCandis("-")

	// get z0 and dz string values
	z0 := a.GetParamValue("z0")
	sdz := a.GetParamValue("dz")

	// get z values -- assumed to be in meters ASL
	z := a.GetDimData("z")
	dz := z[1] - z[0]
	dh := 0.5*dz

	// get fields
	pres := a.GetFieldData("pres")
	temp := a.GetFieldData("temp")
	mr := a.GetFieldData("mr")
	u := a.GetFieldData("u")
	v := a.GetFieldData("v")

	// define some values
	nz := len(z)
	obtop := int(Obtop/dz)
	itrop := int(Ztrop/dz)

	// sanity check on obtop
	if obtop > itrop {
		obtop = itrop
	}
	if obtop < 1 {
		obtop = 1
	}

	// check if data are good at obtop
	datacheck := 1.0e6
	if pres[obtop] > datacheck ||
		temp[obtop] > datacheck ||
		mr[obtop] > datacheck ||
		u[obtop] > datacheck ||
		v[obtop] > datacheck {
		fmt.Fprintln(os.Stderr,"gosounding: bad data at obtop")
		os.Exit(1)
	}

	// convert fields to model units and subtract domain motion
	// from wind profiles
	for i := 0; i < nz; i++ {
		pres[i] = pres[i]*100.0      // Pa
		temp[i] = temp[i] + 273.15   // K
		mr[i] = mr[i]/1000.0         // kg/kg
		u[i] = u[i] - vxsys[0]       // m/s rel to moving frame
		v[i] = v[i] - vysys[0]
	}

	// start constant temperature at tropopause and linearly
	// extrapolate to this level from obtop
	// assume a relative humidity of 50% above obtop
	// also do an approximate hydrostatic calculation above obtop
	for i := obtop + 1; i < nz; i++ {
		if i > itrop {
			temp[i] = temp[itrop]
		} else {
			temp[i] = 2.0* temp[i - 1] - temp[i - 2]
		}
		tstar := 0.5*(temp[i - 1] + temp[i])
		pstar := pres[i - 1] - 0.5*Gee*dz*pres[i - 1]/(thermo.R*tstar)
		pres[i] = pres[i - 1] - Gee*dz*pstar/(thermo.R*tstar)

		// for rhod at high altitudes, we can initially approximate
		// mr by zero
		rhod := thermo.Rho(temp[i],pres[i],0.0)
		mr[i] = 0.5*thermo.Rsat(temp[i],rhod)

		// for winds, just assume a constant value equal to
		// obtop value at higher levels
		u[i] = u[obtop]
		v[i] = v[obtop]
	}

	// compute accurate hydrostatic profile, recomputing pressure
	rho := make([]float64,nz)
	drt := make([]float64,nz)
	ddryent := make([]float64,nz)
	dent := make([]float64,nz)
	dsatent := make([]float64,nz)
	rho[0] = thermo.Rho(temp[0],pres[0],mr[0])
	for i := 1; i < nz; i++ {
		pstar := pres[i - 1] - Gee*rho[i - 1]*(1.0 + mr[i - 1])*dh
		mstar := 0.5*(mr[i - 1] + mr[i])
		tstar := 0.5*(temp[i - 1] + temp[i])
		rhostar := thermo.Rho(tstar,pstar,mstar)
		pres[i] = pres[i - 1] - Gee*rhostar*(1.0 + mstar)*dz
		rho[i] = thermo.Rho(temp[i],pres[i],mr[i])
	}

	// momentum density variables
	ux := make([]float64,nz)
	uy := make([]float64,nz)

	// compute density quantities and print
	for i := 0; i < nz; i++ {
		drt[i] = mr[i]*rho[i]
		ddryent[i] = thermo.Dent(temp[i],rho[i],mr[i])*rho[i]
		dent[i] = thermo.Ent(temp[i],rho[i],mr[i],0.0)*rho[i]
		dsatent[i] = thermo.Sent(temp[i],rho[i])*rho[i]
		ux[i] = u[i]*rho[i]
		uy[i] = v[i]*rho[i]
		fmt.Fprintf(os.Stderr,
			"%3d %8.0f %8.1f %8.3f %8.1f %8.5f %8.1f %8.1f\n",
			i,z[i],pres[i],rho[i],temp[i],mr[i],ux[i],uy[i])
	}

	// create output to stdout
	b := gocandis.NullCandis()
	b.AddParam("z0",z0)
	b.AddParam("dz",sdz)
	b.AddDim("z",z)
	b.AddVField("sst",[]string{},sst)
	b.AddVField("vxsys",[]string{},vxsys)
	b.AddVField("vysys",[]string{},vysys)
	b.AddVField("pres",[]string{"z"},pres)
	b.AddVField("temp",[]string{"z"},temp)
	b.AddVField("rho",[]string{"z"},rho)
	b.AddVField("drt",[]string{"z"},drt)
	b.AddVField("ddryent",[]string{"z"},ddryent)
	b.AddVField("dent",[]string{"z"},dent)
	b.AddVField("dsatent",[]string{"z"},dsatent)
	b.AddVField("ux",[]string{"z"},ux)
	b.AddVField("uy",[]string{"z"},uy)
	b.PutCandis("-")

	// clean up
	a.CloseCandis()
	b.CloseCandis()
}
