// thermo -- do thermodynamic calculations
package thermo

import (
	"math"
	"fmt"
	"os"
)

const (
	Cp = 1005.0   // specific heat const p (J/kg/K)
	Cv = Cp - R   // specific heat const rho (J/kg/K)
	R = 287.05    // dry air gas const (J/kg/K)
	Cpv = 1850.0  // specific heat vapor const p (J/kg/K)
	Rv = 461.50   // vapor gas constant (J/kg/K)
	Cvv = Cpv - Rv // specific heat vapor const rho (J/kg/K)
	Cl = 4218.0   // specific heat liquid water (J/kg/K)
	Ci = 1959.0   // specific heat icd (J/kg/K)
	Gamma = Cp/(Cp - R)
	Kappa = (1.0 - 1.0/Gamma)
	Tfreez = 273.15  // freezing temperature (K)
	Tref = Tfreez  // reference temperature (K)
	Pref = 1.0e5  // reference pressure (Pa)
	Rhoref = Pref/(R*Tref)  // reference density (kg/m^3)
	Epsilon = 0.622  // vapor pressure to mixing ratio constant
	Lfr = 2.5008e6   // latent heat of condensation at 0 C (J/kg)
	Lf = 3.34e5      // latent heat of freezing (J/kg)

	// constants for saturation vapor pressure equation
	Esf = 611.0   // sat vapor pres at 0 C (Pa)
	Mubl = 3.15e6 // binding energy for liquid water (J/kg)
	Mubi = 2.86e6 // binding energy for ice (J/kg)

	// constants for table of saturation vapor pressure
	ntemp = 6901
	temp0 = 10.0
	dtemp = 0.1
	temp1 = temp0 + dtemp*float64(ntemp - 1)

	// constants for table of temp2(rho,ent)
	rho0 = 0.05
	rho1 = 1.3
	ent0 = 10.0
	ent1 = 600.0
	rt0 = 0.0
	rt1 = 0.06
	nrho = 201
	nent = 201
	nrt = 61
	drho = (rho1 - rho0)/float64(nrho - 1)
	dent = (ent1 - ent0)/float64(nent - 1)
	drt = (rt1 - rt0)/float64(nrt - 1)

	// math constants
	Pi = 3.14159
)
var estable [ntemp]float64
var t2tablel [nrho*nent*nrt]float64  // liquid condensate
var t2tablei [nrho*nent*nrt]float64  // ice condensate
var gice int

// initialize thermodynamic tables
func Init(ice int) {

	// make ice global
	gice = ice

	// generate es table
	for itemp := 0; itemp < ntemp; itemp++ {
		temp := temp0 + dtemp*float64(itemp)
		A := Mubl/Rv
		B := (Cl - Cpv)/Rv
		if temp < Tfreez && ice > 0 {
			A = Mubi/Rv
			B = (Ci - Cpv)/Rv
		}

		estable[itemp] = Esf*math.Pow(Tfreez/temp,B)*
			math.Exp(A/Tfreez - A/temp)
	}

	// generate t2l table for saturated case
	for irho := 0; irho < nrho; irho++ {
		for ient := 0; ient < nent; ient++ {
			for irt := 0; irt < nrt; irt++ {
				rho := rho0 + drho*float64(irho)
				ent := ent0 + dent*float64(ient)
				rt := rt0 + drt*float64(irt)
				
				// divide and conquer to narrow the range
				tlo := temp0 + 1.0
				thi := temp1 - 1.0
				qlo := qcalcl(tlo,rho,ent,rt)
				qhi := qcalcl(thi,rho,ent,rt)
				if qlo >= 0.0 || qhi <= 0.0 {
					fmt.Println("thermo.Init: range error,qlo,qhi,rho,ent,rt:",qlo,qhi,rho,ent,rt)
					os.Exit(1)
				}
				for i := 0; i < 5; i++ {
					tmid := 0.5*(tlo + thi)
					qmid := qcalcl(tmid,rho,ent,rt)
					if qmid > 0 {
						thi = tmid
						qhi = qmid
					} else {
						tlo = tmid
						qlo = qmid
					}
				}
				
				// now refine with Newton's method -- always
				// completes in 3 or 4 cycles
				t := 0.5*(tlo + thi)
				for i := 0; i < 10; i++ {
					q1 := qcalcl(t,rho,ent,rt)
					q2 := qcalcl(t + 1,rho,ent,rt)
					tp := -q1/(q2 - q1)
					t = t + tp
					if math.Abs(tp) < 1.e-3 {break}
				}
				t2tablel[K(irho,ient,irt)] = t
			}
		}
	}

	// generate t2i table for saturated case (only if ice)
	if gice != 0 {
		for irho := 0; irho < nrho; irho++ {
			for ient := 0; ient < nent; ient++ {
				for irt := 0; irt < nrt; irt++ {
					rho := rho0 + drho*float64(irho)
					ent := ent0 + dent*float64(ient)
					rt := rt0 + drt*float64(irt)
					
					// divide and conquer to narrow the range
					tlo := temp0 + 1.0
					thi := temp1 - 1.0
					qlo := qcalci(tlo,rho,ent,rt)
					qhi := qcalci(thi,rho,ent,rt)
					if qlo >= 0.0 || qhi <= 0.0 {
						fmt.Println("thermo.Init: range error,qlo,qhi,rho,ent,rt:",qlo,qhi,rho,ent,rt)
						os.Exit(1)
					}
					for i := 0; i < 5; i++ {
						tmid := 0.5*(tlo + thi)
						qmid := qcalci(tmid,rho,ent,rt)
						if qmid > 0 {
							thi = tmid
							qhi = qmid
						} else {
							tlo = tmid
							qlo = qmid
						}
					}
					
					// now refine with Newton's method -- always
					// completes in 3 or 4 cycles
					t := 0.5*(tlo + thi)
					for i := 0; i < 10; i++ {
						q1 := qcalci(t,rho,ent,rt)
						q2 := qcalci(t + 1,rho,ent,rt)
						tp := -q1/(q2 - q1)
						t = t + tp
						if math.Abs(tp) < 1.e-3 {break}
					}
					t2tablei[K(irho,ient,irt)] = t
				}
			}
		}
	}
}

func qcalcl(t, rho, ent, rt float64) (q float64) {
	q = Entl(t,rho,rt,0.0) - ent
	return
}

func qcalci(t, rho, ent, rt float64) (q float64) {
	q = Enti(t,rho,rt,0.0) - ent
	return
}

// compute the saturation vapor pressure (Pa) from temperature (K)
// theoretical formula used
func Es(td float64) (es float64) {

	// get es from table (temperature should remain in limits)
	if td < temp0 || td > temp1 {
		fmt.Println("thermo.Es:",td,"outside temp limits")
		os.Exit(1)
	}
	tindex := int((td - temp0)/dtemp)
	if tindex < 0 || tindex > ntemp - 2 {
		fmt.Println("tindex,td:",tindex,td)
		os.Exit(1)
	}
	tfrac := (td - (temp0 + dtemp*float64(tindex)))/dtemp
	es = estable[tindex]*(1.0 - tfrac) + estable[tindex + 1]*tfrac
	return
}

// compute saturation mixing ratio (g/g) as a function of
// temperature (K) and density of dry air (kg/m^3)
func Rsat(temp, rho float64) (rsat float64) {
	es := Es(temp)
	rsat = Epsilon*es/Presd(rho,temp)
	return
}

// compute the vapor mixing ratio (g/g) from temperature (K), dry air
// density (kg/m^3), and total cloud water mixing ratio (g/g)
func Mr(temp, rho, rt float64) (r float64) {
	mrsat := Rsat(temp,rho)
	if rt > mrsat {
		r = mrsat
	} else {
		r = rt
	}
	return
}

// compute the dry entropy (J/kg/K) from temp (K), dry density (kg/m^3)
func Dent(temp, rho, rt float64) (dent float64) {
	rv := Mr(temp,rho,rt)
	logrvterm := 0.0
	if rv > 1.e-10 {
		logrvterm = rv*Rv*math.Log(Pref*rv/(Epsilon*Esf))
	}
	dent = (Cv + rv*Cvv)*math.Log(temp/Tref) -
		(R + rv*Rv)*math.Log(rho/Rhoref) -
		logrvterm
	return
}

// compute approximate dry entropy (J/kg/K) from temp (K), pressure (Pa)
func Dent2(temp, p float64) (dent2 float64) {
	dent2 = Cp*math.Log(temp/Tref) - R*math.Log(p/Pref)
	return
}

// compute the entropy (J/kg/K) of precip from temp(K) and
// precip mixing ratio
func Entrr(temp, rr float64) (entrr float64) {

	// select constants by temperature with smooth transition
	Cx := Cl
	Lx := 0.0
	if gice != 0 {
		Cx = Ftrans(Cl,Ci,temp)
		Lx = Ftrans(0.0,Lf,temp)
	}

	// calculate entropy
	entrr = rr*(Cx*math.Log(temp/Tref) - Lx/Tfreez)
	return
}

// compute the moist entropy (J/kg/K) from temp (K), dry density (kg/m^3),
// advected water substance and precipitation mixing ratios (g/g)
// above freezing version
func Entl(temp, rho, rt, rr float64) (ent float64) {

	// water physics
	Cx := Cl
	Lx := 0.0

	// avoid singularity of 0*ln(0)
	rv := Mr(temp,rho,rt)
	rc := rt - rv
	rx := rc + rr
	logrvterm := 0.0
	if rv > 1.e-10 {
		logrvterm = rv*Rv*math.Log(Pref*rv/(Epsilon*Esf))
	}
	ent = (Cv + rv*Cvv + rx*Cx)*math.Log(temp/Tref) -
		(R + rv*Rv)*math.Log(rho/Rhoref) -
		logrvterm +
		(Lfr*rv - Lx*rx)/Tfreez
	return
}

// compute the moist entropy (J/kg/K) from temp (K), dry density (kg/m^3),
// advected water substance and precipitation mixing ratios (g/g)
// below freezing version
func Enti(temp, rho, rt, rr float64) (ent float64) {

	// ice physics
	Cx := Ci
	Lx := Lf

	// avoid singularity of 0*ln(0)
	rv := Mr(temp,rho,rt)
	rc := rt - rv
	rx := rc + rr
	logrvterm := 0.0
	if rv > 1.e-10 {
		logrvterm = rv*Rv*math.Log(Pref*rv/(Epsilon*Esf))
	}
	ent = (Cv + rv*Cvv + rx*Cx)*math.Log(temp/Tref) -
		(R + rv*Rv)*math.Log(rho/Rhoref) -
		logrvterm +
		(Lfr*rv - Lx*rx)/Tfreez
	return
}

// compute correct entropy based on temperature and ice flag
func Ent(temp, rho, rt, rr float64) (ent float64) {
	entl := Entl(temp,rho,rt,rr)
	if gice != 0 {
		enti := Enti(temp,rho,rt,rr)
		ent = Ftrans(entl,enti,temp)
	} else {
		ent = entl
	}
	return
}

// calculate saturated entropy as a function of temp (K)
// and dry density (kg/m^3) -- set rc and rr to zero
func Sent(temp, rho float64) (sent float64) {
	sent = Entl(temp, rho, Rsat(temp,rho),0.0)
	return
}

// compute dry air density (kg/m^3) from temperature (K), pressure (Pa),
// and total cloud water mixing ratio (g/g)
// IMPORTANT -- assumes rv = rt, i.e., air is unsaturated
func Rho(temp, p, rt float64) (rho float64) {
	rho = p/(R*temp*(1.0 + rt/Epsilon))
	return
}

// compute partial pressure of dry air (Pa) from dry density (kg/m^3) and
// temperature (K)
func Presd(rho, temp float64) (p float64) {
	p = R*temp*rho
	return
}

// compute total pressure (Pa) from dry air density (kg/m^3), total
// cloud water (g/g), and moist entropy (J/kg/K)
func Pres2(rho, rt, rr, ent float64) (p float64) {

	// approximate the temperature using total entropy
	temp := Temp2(rho,rt,ent)

	// compute the precipitation entropy based on this and
	// recompute the temperature
	entrr := Entrr(temp,rr)

	// recompute the temperature using the advected entropy
	temp = Temp2(rho,rt,ent - entrr)
	
	// compute the pressure using the refined temperature
	pd := Presd(rho,temp)
	pv := pd*Mr(temp,rho,rt)/Epsilon
	p = pd + pv
	return
}

// compute temperature (K) from density (kg/m^3), pressure (Pa),
// and total water mixing ratio
func Temp(rho, p, rt float64) (temp float64) {
	oldtemp := 0.0
	temp = p/(R*rho)
	rv := Mr(temp,rho,rt)
	for i := 0; i < 20; i++ {
		oldfrac := float64(i)/20.0
		pd := p/(1.0 + rv/Epsilon)
		oldtemp = temp
		temp = (1.0 - oldfrac)*pd/(R*rho) + oldfrac*oldtemp
		if math.Abs(temp - oldtemp) < 0.01 {break}
		rv = Mr(temp,rho,rt)
	}
	return
}

// compute temperature (K) from density (Kg/m^3), total cloud water
// mixing ratio, and advected moist entropy (J/kg/K)
// liquid case
func Temp2l(rho, rt, ent float64) (temp float64) {

	// test for saturation or low density/high altitude
	logrvterm:= 0.0   // avoid singular case of 0*log(0)
	if rt > 1.e-10 {
		logrvterm = rt*Rv*math.Log(Pref*rt/(Epsilon*Esf))
	}
	x := (ent + (R + rt*Rv)*math.Log(rho/Rhoref) +
		logrvterm - Lfr*rt/Tfreez)/(Cv + Cvv*rt)
	t := Tref*math.Exp(x)
	rs := Rsat(t,rho)
	if rt < rs {

		// unsaturated -- we are done
		return(t)
	} else {

		// saturated -- get temp from table
		xrho := (rho - rho0)/drho
		irho := int(xrho)
		if irho < 0 || irho > nrho - 2 {
			fmt.Println("thermo.Temp2: range error, rho:",rho)
			os.Exit(1)
		}
		xent := (ent - ent0)/dent
		ient := int(xent)
		if ient < 0 || ient > nent - 2 {
			fmt.Println("thermo.Temp2: range error, ent:",ent)
			os.Exit(1)
		}
		xrt := (rt - rt0)/drt
		irt := int(xrt)
		if irt < 0 || irt > nrt - 2 {
			fmt.Println("thermo.Temp2: range error, rt:",rt)
			os.Exit(1)
		}
		jrho := irho + 1
		jent := ient + 1
		jrt := irt + 1
		xrho -= float64(irho)
		xent -= float64(ient)
		xrt -= float64(irt)
		yrho := 1.0 - xrho
		yent := 1.0 - xent
		yrt := 1.0 - xrt
		t2 := t2tablel[K(jrho,jent,jrt)]*xrho*xent*xrt +
			t2tablel[K(irho,jent,jrt)]*yrho*xent*xrt +
			t2tablel[K(jrho,ient,jrt)]*xrho*yent*xrt +
			t2tablel[K(irho,ient,jrt)]*yrho*yent*xrt +
			t2tablel[K(jrho,jent,irt)]*xrho*xent*yrt +
			t2tablel[K(irho,jent,irt)]*yrho*xent*yrt +
			t2tablel[K(jrho,ient,irt)]*xrho*yent*yrt +
			t2tablel[K(irho,ient,irt)]*yrho*yent*yrt
		return(t2)
	}
}

// compute temperature (K) from density (Kg/m^3), total water
// mixing ratio, and advected moist entropy (J/kg/K)
// ice case
func Temp2i(rho, rt, ent float64) (temp float64) {

	// test for saturation or low density/high altitude
	logrvterm:= 0.0   // avoid singular case of 0*log(0)
	if rt > 1.e-10 {
		logrvterm = rt*Rv*math.Log(Pref*rt/(Epsilon*Esf))
	}
	x := (ent + (R + rt*Rv)*math.Log(rho/Rhoref) +
		logrvterm - Lfr*rt/Tfreez)/(Cv + Cvv*rt)
	t := Tref*math.Exp(x)
	rs := Rsat(t,rho)
	if rt < rs {

		// unsaturated -- we are done
		return(t)
	} else {

		// saturated -- get temp from table
		xrho := (rho - rho0)/drho
		irho := int(xrho)
		if irho < 0 || irho > nrho - 2 {
			fmt.Println("thermo.Temp2: range error, rho:",rho)
			os.Exit(1)
		}
		xent := (ent - ent0)/dent
		ient := int(xent)
		if ient < 0 || ient > nent - 2 {
			fmt.Println("thermo.Temp2: range error, ent:",ent)
			os.Exit(1)
		}
		xrt := (rt - rt0)/drt
		irt := int(xrt)
		if irt < 0 || irt > nrt - 2 {
			fmt.Println("thermo.Temp2: range error, rt:",rt)
			os.Exit(1)
		}
		jrho := irho + 1
		jent := ient + 1
		jrt := irt + 1
		xrho -= float64(irho)
		xent -= float64(ient)
		xrt -= float64(irt)
		yrho := 1.0 - xrho
		yent := 1.0 - xent
		yrt := 1.0 - xrt
		t2 := t2tablei[K(jrho,jent,jrt)]*xrho*xent*xrt +
			t2tablei[K(irho,jent,jrt)]*yrho*xent*xrt +
			t2tablei[K(jrho,ient,jrt)]*xrho*yent*xrt +
			t2tablei[K(irho,ient,jrt)]*yrho*yent*xrt +
			t2tablei[K(jrho,jent,irt)]*xrho*xent*yrt +
			t2tablei[K(irho,jent,irt)]*yrho*xent*yrt +
			t2tablei[K(jrho,ient,irt)]*xrho*yent*yrt +
			t2tablei[K(irho,ient,irt)]*yrho*yent*yrt
		return(t2)
	}
}

// compute proper temperature from density, total cloud water, and
// total moist entropy
func Temp2(rho, rt, ent float64) (temp float64) {
	if gice == 0 {
		temp = Temp2l(rho,rt,ent)
		return
	} else {
		temp2l := Temp2l(rho,rt,ent)
		temp2i := Temp2i(rho,rt,ent)

		// various cases
		if temp2l > Tfreez && temp2i > Tfreez {
			temp = temp2l
			return
		} else if temp2l < Tfreez && temp2i < Tfreez {
			temp = temp2i
			return
		} else { // mixed phase at freezing
			temp = Tfreez
			return
		}
	}
}

// make transition across freezing level
func Ftrans(xwarm, xcold, temp float64) (xtrans float64) {

	// step transition
	if temp < Tfreez {
		xtrans = xcold
	} else {
		xtrans = xwarm
	}
	return
}

func K(irho, ient, irt int) (k int) {
	k = irho + nrho*(ient + nent*irt)
	return
}
