// radiation -- Interactive radiation package -- port of C language
// toyrad package.  Include only the MEGACHANNEL option
package toyrad

import (
	"math"
	"thermo"
)

const (
	Co2 = 0
	Cont = 1
	Rtref = 0.02
	Sbconst = 5.67e-8
	Solarconst = 1360.0
	Nbands = 13
	E = 2.7182818284590451
	Etablesize = 100001
)

// table for lookup of safeexp values
var table [Etablesize]float64
var begin bool = true

// Compute net radiative entropy source term sent and the solar flux
// are all cell-centered -- sent and sflux are returned variables
func Radiation(mz, sun int, cld, time, solarabs, solarwv, stemp,
	dz float64, rt, p, rho, rsat, sent, sflux, flux []float64) {
	
	// fraction of Planck-weighted spectrum occupied by each band
	fract := []float64{0.11,0.20,0.13,0.11,0.09,0.08,0.07,0.06,
		0.05,0.04,0.03,0.02,0.01}

	// specific absorption coefficents for each band
	kappa := []float64{0.002,0.0005,0.001,0.00316,0.01,0.0316,0.1,0.316,
		1.0,3.16,10.0,31.6,100.0}

	// compute cell-centered temperatures and define sigma*T^4
	temp := make([]float64,mz)
	bprof := make([]float64,mz)
	for iz := 0; iz < mz; iz++ {
		temp[iz] = thermo.Temp(rho[iz],p[iz],rt[iz])
		mytemp := temp[iz]
		bprof[iz] = Sbconst*mytemp*mytemp*mytemp*mytemp
	}

	// define the surface Planck function
	bsurf := Sbconst*stemp*stemp*stemp*stemp

	// compute the cosine of solar zenith angle
	scos := 0.0
	if sun == 0 {  // sun is turned off
		scos = 0.0
	} else if sun == 1 {  // diurnal average scos
		scos = 1.0/3.14159
	} else {  // time-dependent scos
		zangle := 6.28318*time/86400.0
		scos = math.Cos(zangle)
		if scos < 0.0 {
			scos = 0.0
		}
	}

	// define some cell-edge auxiliary slices
	netflux := make([]float64,mz + 1)
	efluxup := make([]float64,mz + 1)
	efluxdn := make([]float64,mz + 1)
	solflux := make([]float64,mz + 1)
	tau := make([]float64,mz + 1)
	etemp := make([]float64,mz + 1)
	sminus := make([]float64,mz + 1)
	splus := make([]float64,mz + 1)

	// compute cell-edge temperatures
	for iz := 0; iz <= mz; iz++ {
		ep := 0.0
		erho := 0.0
		ert := 0.0
		if iz == 0 {
			ep = 1.5*p[0] - 0.5*p[1]
			erho = 1.5*rho[0] - 0.5*rho[1]
			ert = 1.5*rt[0] - 0.5*rt[1]
		} else if iz == mz {
			ep = 1.5*p[mz - 1] - 0.5*p[mz - 2]
			erho = 1.5*rho[mz - 1] - 0.5*rho[mz - 2]
			ert = 1.5*rt[mz - 1] - 0.5*rt[mz - 2]
		} else {
			ep = 0.5*(p[iz - 1] + p[iz])
			erho = 0.5*(rho[iz - 1] + rho[iz])
			ert = 0.5*(rt[iz - 1] + rt[iz])
		}
		etemp[iz] = thermo.Temp(erho,ep,ert)
	}

	// zero edge flux array
	for iz := 0; iz <= mz; iz++ {
		netflux[iz] = 0.0
		efluxup[iz] = 0.0
		efluxdn[iz] = 0.0
	}

	// loop on the thermal IR bands
	for ib := 0; ib < Nbands; ib++ {

		// compute tau, the actual absorption coefficient, for this band
		taucomp(mz,ib,dz,rt,rsat,rho,etemp,kappa[ib],cld,tau)

		// add in the upward flux
		iplus := 0.0
		for jz := 0; jz <= mz; jz++ {
			if jz == 0 {
				iplus = fract[ib]*bsurf
			} else {
				dtau := tau[jz] - tau[jz - 1]
				efact := safeexp(elimit(-dtau))
				iplus = iplus*efact +
					(1.0 - efact)*fract[ib]*bprof[jz - 1]
			}
			efluxup[jz] += iplus
		}

		// subtract off the downward flux -- zero at top
		iminus := 0.0
		for jz := mz; jz >= 0; jz-- {
			if jz == mz {
				iminus = 0.0
			} else {
				dtau := tau[jz + 1] - tau[jz]
				efact := safeexp(elimit(-dtau))
				iminus = iminus*efact +
					(1.0 - efact)*fract[ib]*bprof[jz]
			}
			efluxdn[jz] += iminus
		}
	}

	// different treatment for solar visible and near-IR
	// integrate unnormalized iplus and iminus up from surface
	// assuming that iplus = 0 at surface
	// if scos == 0 there is no solar contribution
	if scos == 0.0 {
		for iz := 0; iz <= mz; iz++ {
			splus[iz] = 0.0
			sminus[iz] = 0.0
		}
	} else {
		sminus[0] = 1.0
		splus[0] = 0.0
		for iz := 1; iz <= mz; iz++ {
			jz := iz - 1
			
			// compute cloud water and water vapor
			rl := rt[jz] - rsat[jz]
			if rl < 0.0 {
				rl = 0.0
			}
			rvap := rt[jz] - rl

			// scattering is done just by cloud water
			alpha := rho[jz]*cld*rl
			
			// sanity check
			if alpha > 0.5/dz {
				alpha = 0.5/dz
			}

			// estimate solar absorption as
			// fraction of scattering --
			// water vapor continuum also contributes
			// to solar absorption
			gamma := solarabs*alpha + rho[jz]*solarwv*rvap

			// sanity check
			if gamma > 0.5/dz {
				gamma = 0.5/dz
			}

			// calc transfer in scattering
			dntoup := alpha*sminus[jz]*dz
			uptodn := alpha*splus[jz]*dz
			splus[iz] = splus[jz] + dntoup - uptodn
			sminus[iz] = sminus[jz] - uptodn + dntoup

			// sanity check
			if sminus[iz] < 0.0 {sminus[iz] = 0.0}
			if splus[iz] < 0.0 {splus[iz] = 0.0}

			// calc absorption
			splus[iz] = splus[iz]*(1.0 - gamma*dz)
			sminus[iz] = sminus[iz]*(1.0 + gamma*dz)

			// sanity check
			if sminus[iz] < 0.0 {sminus[iz] = 0.0}
			if splus[iz] < 0.0 {splus[iz] = 0.0}
		}

		// normalize splus and sminus and add to total flux
		if sminus[mz] > 0.0 {
			ratio := scos*Solarconst/sminus[mz]
			for iz := 0; iz <= mz; iz++ {
				splus[iz] *= ratio
				efluxup[iz] += splus[iz]
				sminus[iz] *= ratio
				efluxdn[iz] += sminus[iz]
			}
		}
	}

	// compute the net IR + solar eflux and separate solar flux
	for iz := 0; iz <= mz; iz++ {
		netflux[iz] = efluxup[iz] - efluxdn[iz]
		solflux[iz] = splus[iz] - sminus[iz]
	}

	// compute the cell-centered quantities
	for iz := 0; iz < mz; iz++ {
		jz := iz + 1

		// convert from energy/vol/time to entropy/vol/time
		normfactor := 1.0/temp[iz]

		// compute the flux divergence to get net entropy source
		sent[iz] = -normfactor*(netflux[jz] - netflux[iz])/dz

		// interpolate the net solar flux
		sflux[iz] = (solflux[iz] + solflux[jz])/2.0

		// interpolate the net total flux
		flux[iz] = (netflux[iz] + netflux[jz])/2.0
	}
}
	
// taucomp -- compute the path length for a band -- tau is defined on
// cell edges and is zero at the surface -- this is for IR
func taucomp(mz, ib int, dz float64, rt, rsat, rho, etemp []float64,
	kappa, cld float64, tau []float64) {

	// integrate upward from surface to get tau -- iz = upper edge index,
	// kz = lower edge index and cell-centered index
	tau[0] = 0.0
	for iz := 1; iz <= mz; iz++ {
		kz := iz - 1
		
		// compute pressure broadening factor
		temp := 0.5*(etemp[iz] + etemp[kz])
		pbroad := math.Pow(temp/thermo.Tref,0.5)*rho[kz]/thermo.Rhoref
		
		// compute vapor-condensate split
		rl := rt[kz] - rsat[kz]
		if rl < 0.0 {
			rl = 0.0
		}
		rvap := rt[kz] - rl
		rfract := rvap/Rtref
		
		// compute mutot in cell center for desired band
		mutot := 0.0
		if ib == Co2 {                 // carbon dioxide band
			mutot = kappa*pbroad
		} else if ib == Cont {           // water vapor continuum
			mutot = kappa*rfract*rfract			
		} else {                           // water vapor bands
			mutot = kappa*pbroad*rfract
		}

		// add in cloud part
		mutot += cld*rl

		// compute tau at the next level
		tau[iz] = tau[kz] + rho[kz]*mutot*dz
	}
}

// safeexp -- Do our own computation of the exponential function --
// this results from the sporadic long times required to compute
// the exponential function using the standard GNU math library on
// 64 bit systems (AMD, Intel)
func safeexp(x float64) (result float64) {

	// fill the lookup table once
	if begin {
		for i := 0; i < Etablesize; i++ {
			table[i] = math.Exp(float64(i)/float64(Etablesize))
		}
		begin = false
	}

	xx := x
	result = 1.0
	for xx >= 0.0 {
		xx -= 1.0
		result *= E
	}
	for xx < 0.0 {
		xx += 1.0
		result /= E
	}
	i := int(xx*float64(Etablesize))
	if i < 0 {
		i = 0
	}
	if i >= Etablesize {
		i = Etablesize - 1
	}
	result *= table[i]
	return(result)
}

// elimit -- limit the value to +- 75 to avoid blowing up exponential fn
func elimit(val float64) (lval float64) {
	lval = val
	if lval > 75.0 {lval = 75.0}
	if lval < -75.0 {lval = -75.0}
	return(lval)
}
